from unittest.mock import patch, Mock
import pytest
import requests
from spark_expectations.core.exceptions import SparkExpectationsPagerDutyException
from spark_expectations.notifications.plugins.pagerduty import SparkExpectationsPagerDutyPluginImpl


@patch("spark_expectations.notifications.plugins.pagerduty.SparkExpectationsContext", autospec=True)
def test_send_incident_success(_mock_context):
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    _mock_context.get_enable_pagerduty = True
    _mock_context.get_pagerduty_webhook_url = "https://events.pagerduty.com/v2/change/enqueue"
    _mock_context.get_pagerduty_integration_key = "test_integration_key"
    _mock_context.product_id = "test_product"
    _mock_context.get_table_name = "test_table"

    _config_args = {"message": "Spark expectations job has been failed"}

    # Mock requests.post to return a response with status code 200
    with patch.object(requests, "post") as mock_post:
        mock_response = Mock()
        mock_response.status_code = 202
        mock_post.return_value = mock_response

        print("TEST")
        print(_mock_context.get_pagerduty_webhook_url)
        print(_mock_context.get_pagerduty_integration_key)
        pagerduty_handler.send_notification(_context=_mock_context, _config_args=_config_args)
        mock_post.assert_called_once_with(
            _mock_context.get_pagerduty_webhook_url,
            json={
                "routing_key": _mock_context.get_pagerduty_integration_key,
                "dedup_key": "spark_expectations_test_product_test_table_failure",
                "event_action": "trigger",
                "payload": {
                    "summary": "Spark expectations job has been failed",
                    "source": "Spark Expectations",
                    "severity": "error",
                },
            },
            headers={"Content-Type": "application/json"},
            timeout=10,
        )


@patch("spark_expectations.notifications.plugins.pagerduty.SparkExpectationsContext", autospec=True, spec_set=True)
def test_send_incident_exception(_mock_context):
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    _mock_context.get_enable_pagerduty = True
    _mock_context.get_pagerduty_webhook_url = "http://test_webhook_url"
    _config_args = {"message": "Spark expectations job has been failed"}

    with patch.object(requests, "post") as mock_post:
        mock_response = Mock()
        mock_response.status_code = 404
        mock_post.return_value = mock_response

        with pytest.raises(SparkExpectationsPagerDutyException):
            pagerduty_handler.send_notification(_context=_mock_context, _config_args=_config_args)


@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend", autospec=True, spec_set=True)
@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend.get_secret", autospec=True, spec_set=True)
def test_get_cerberus_integration_key(_mock_secret_handler, _mock_get_secret):
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    pd_secret_dict = {
        "se.streaming.secret.type": "cerberus",
        "se.streaming.cerberus.url": "http://cerberus.url",
        "se.streaming.cerberus.sdb.path": "/path/to/secret",
        "spark.expectations.notifications.pagerduty.integration.key": "integration_key",
    }
    _mock_get_secret.return_value = {"integration_key": "test_integration_key"}
    _mock_secret_handler.get_secret = _mock_get_secret

    integration_key = pagerduty_handler._get_cerberus_integration_key(_mock_secret_handler, pd_secret_dict)
    assert integration_key == "test_integration_key"


@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend", autospec=True, spec_set=True)
@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend.get_secret", autospec=True, spec_set=True)
def test_get_cerberus_integration_key_none(_mock_secret_handler, _mock_get_secret):
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    pd_secret_dict = {
        "se.streaming.secret.type": "cerberus",
        "se.streaming.cerberus.url": "http://cerberus.url",
        "se.streaming.cerberus.sdb.path": "/path/to/secret",
        "spark.expectations.notifications.pagerduty.integration.key": None,
    }
    _mock_get_secret.return_value = None
    _mock_secret_handler.get_secret = _mock_get_secret

    integration_key = pagerduty_handler._get_cerberus_integration_key(_mock_secret_handler, pd_secret_dict)
    assert integration_key is None


@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend", autospec=True, spec_set=True)
@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend.get_secret", autospec=True, spec_set=True)
def test_get_databricks_integration_key(_mock_secret_handler, _mock_get_secret):
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    pd_secret_dict = {
        "se.streaming.secret.type": "databricks",
        "se.streaming.databricks.url": "http://databricks.url",
        "se.streaming.databricks.secret.scope": "my_secret_scope",
        "spark.expectations.notifications.pagerduty.integration.key": "integration_key",
    }
    _mock_get_secret.return_value = "test_integration_key"
    _mock_secret_handler.get_secret = _mock_get_secret

    integration_key = pagerduty_handler._get_databricks_integration_key(_mock_secret_handler, pd_secret_dict)
    assert integration_key == "test_integration_key"


@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend", autospec=True, spec_set=True)
@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend.get_secret", autospec=True, spec_set=True)
def test_get_databricks_integration_key_none(_mock_secret_handler, _mock_get_secret):
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    pd_secret_dict = {
        "se.streaming.secret.type": "databricks",
        "se.streaming.databricks.url": "http://databricks.url",
        "se.streaming.databricks.secret.scope": "my_secret_scope",
        "spark.expectations.notifications.pagerduty.integration.key": None,
    }
    _mock_get_secret.return_value = None
    _mock_secret_handler.get_secret = _mock_get_secret

    integration_key = pagerduty_handler._get_databricks_integration_key(_mock_secret_handler, pd_secret_dict)
    assert integration_key is None

@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend", autospec=True, spec_set=True)
def test_retrieve_integration_key_cerberus(_mock_secret_handler):
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    pd_secret_dict = {
        "se.streaming.secret.type": "cerberus",
        "se.streaming.cerberus.url": "http://cerberus.url",
        "se.streaming.cerberus.sdb.path": "/path/to/secret",
        "spark.expectations.notifications.pagerduty.integration.key": "integration_key",
    }

    with patch("spark_expectations.notifications.plugins.pagerduty.SparkExpectationsPagerDutyPluginImpl._get_cerberus_integration_key", return_value="test_integration_key") as mock_cerberus_integration_key:
        integration_key = pagerduty_handler._retrieve_integration_key(_mock_secret_handler, "cerberus", pd_secret_dict)
        mock_cerberus_integration_key.assert_called_once_with(_mock_secret_handler, pd_secret_dict)
        assert integration_key == "test_integration_key"

@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend", autospec=True, spec_set=True)
def test_retrieve_integration_key_databricks(_mock_secret_handler):
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    pd_secret_dict = {
        "se.streaming.secret.type": "databricks",
        "se.streaming.databricks.url": "http://databricks.url",
        "se.streaming.databricks.secret.scope": "my_secret_scope",
        "spark.expectations.notifications.pagerduty.integration.key": "integration_key",
    }

    with patch("spark_expectations.notifications.plugins.pagerduty.SparkExpectationsPagerDutyPluginImpl._get_databricks_integration_key", return_value="test_integration_key") as mock_databricks_integration_key:
        integration_key = pagerduty_handler._retrieve_integration_key(_mock_secret_handler, "databricks", pd_secret_dict)
        mock_databricks_integration_key.assert_called_once_with(_mock_secret_handler, pd_secret_dict)
        assert integration_key == "test_integration_key"


@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend", autospec=True, spec_set=True)
def test_retrieve_integration_key_none(_mock_secret_handler):
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    pd_secret_dict = {
        "se.streaming.secret.type": "not_a_valid_type",
        "se.streaming.databricks.url": "http://databricks.url",
        "se.streaming.databricks.secret.scope": "my_secret_scope",
        "spark.expectations.notifications.pagerduty.integration.key": "integration_key",
    }

    integration_key = pagerduty_handler._retrieve_integration_key(_mock_secret_handler, "not_a_valid_type", pd_secret_dict)
    assert integration_key is None


@patch("spark_expectations.notifications.plugins.pagerduty.SparkExpectationsContext", autospec=True, spec_set=True)
def test_get_pd_integration_key_method_with_databricks(_mock_context):
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    _mock_context.get_pagerduty_integration_key = None
    _mock_context.get_pagerduty_creds_dict = {
        "se.streaming.secret.type": "databricks",
        "se.streaming.databricks.url": "http://databricks.url",
        "se.streaming.databricks.secret.scope": "my_secret_scope",
        "spark.expectations.notifications.pagerduty.integration.key": "integration_key",
    }

    with patch("spark_expectations.notifications.plugins.pagerduty.SparkExpectationsPagerDutyPluginImpl._retrieve_integration_key", return_value="test_integration_key") as mock_retrieve_integration_key:
        pagerduty_handler._get_pd_integration_key(_mock_context)
        mock_retrieve_integration_key.assert_called_once()


@patch("spark_expectations.notifications.plugins.pagerduty.SparkExpectationsContext", autospec=True, spec_set=True)
def test_get_pd_integration_key_method_with_cerberus(_mock_context):
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    _mock_context.get_pagerduty_integration_key = None
    _mock_context.get_pagerduty_creds_dict = {
        "se.streaming.secret.type": "cerberus",
        "se.streaming.cerberus.url": "http://cerberus.url",
        "se.streaming.cerberus.secret.scope": "my_secret_scope",
        "spark.expectations.notifications.pagerduty.integration.key": "integration_key",
    }

    with patch("spark_expectations.notifications.plugins.pagerduty.SparkExpectationsPagerDutyPluginImpl._retrieve_integration_key", return_value="test_integration_key") as mock_retrieve_integration_key:
        pagerduty_handler._get_pd_integration_key(_mock_context)
        mock_retrieve_integration_key.assert_called_once()

@patch("spark_expectations.notifications.plugins.pagerduty.SparkExpectationsContext", autospec=True, spec_set=True)
def test_get_pd_integration_key_method_missing_key_with_key_error(_mock_context):
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    _mock_context.get_pagerduty_integration_key = None
    _mock_context.get_pagerduty_creds_dict = {
        "se.streaming.secret.type": "databricks",
        "se.streaming.databricks.url": "http://databricks.url",
        "se.streaming.databricks.secret.scope": "my_secret_scope",
        "spark.expectations.notifications.pagerduty.integration.key": None,
    }

    with patch("spark_expectations.notifications.plugins.pagerduty.SparkExpectationsPagerDutyPluginImpl._retrieve_integration_key") as mock_retrieve_integration_key:
        mock_retrieve_integration_key.side_effect = KeyError("KeyError: PagerDuty integration key is missing.")
        with pytest.raises(SparkExpectationsPagerDutyException, match="KeyError: PagerDuty integration key is missing."):
            pagerduty_handler._get_pd_integration_key(_mock_context)


@patch("spark_expectations.notifications.plugins.pagerduty.SparkExpectationsContext", autospec=True, spec_set=True)
def test_get_pd_integration_key_method_failed_retrieval(_mock_context):
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    _mock_context.get_pagerduty_integration_key = None
    _mock_context.get_pagerduty_creds_dict = {
        "se.streaming.secret.type": "databricks",
        "se.streaming.databricks.url": "http://databricks.url",
        "se.streaming.databricks.secret.scope": "my_secret_scope",
        "spark.expectations.notifications.pagerduty.integration.key": "integration_key",
    }

    with patch("spark_expectations.notifications.plugins.pagerduty.SparkExpectationsPagerDutyPluginImpl._retrieve_integration_key", side_effect=Exception("Generic error")):
        with pytest.raises(SparkExpectationsPagerDutyException, match="Failed to retrieve PagerDuty integration key."):
            pagerduty_handler._get_pd_integration_key(_mock_context)

@patch("spark_expectations.notifications.plugins.pagerduty.SparkExpectationsContext", autospec=True, spec_set=True)
def test_get_pd_integration_key_method_missing_key(_mock_context):
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    _mock_context.get_pagerduty_integration_key = None
    _mock_context.get_pagerduty_creds_dict = {
        "se.streaming.secret.type": "databricks",
        "se.streaming.databricks.url": "http://databricks.url",
        "se.streaming.databricks.secret.scope": "my_secret_scope",
        "spark.expectations.notifications.pagerduty.integration.key": None,
    }

    with patch("spark_expectations.notifications.plugins.pagerduty.SparkExpectationsPagerDutyPluginImpl._retrieve_integration_key") as mock_retrieve_integration_key:
        mock_retrieve_integration_key.return_value = None
        with pytest.raises(SparkExpectationsPagerDutyException, match="PagerDuty integration key is not set."):
            pagerduty_handler._get_pd_integration_key(_mock_context)

@patch("spark_expectations.notifications.plugins.pagerduty.SparkExpectationsContext", autospec=True, spec_set=True)
def test_send_incident_pagerduty_disabled(_mock_context):
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    _mock_context.get_enable_pagerduty = False
    _mock_context.get_pagerduty_webhook_url = "http://test_webhook_url"
    _config_args = {"message": "test message"}

    with patch.object(requests, "post") as mock_post:
        pagerduty_handler.send_notification(_context=_mock_context, _config_args=_config_args)
        mock_post.post.assert_not_called()


@patch("spark_expectations.notifications.plugins.pagerduty.SparkExpectationsContext", autospec=True, spec_set=True)
def test_send_incident_skipped_for_non_failure_message(_mock_context):
    """Test that PagerDuty notifications are skipped for non-failure messages like job start/completion"""
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    _mock_context.get_enable_pagerduty = True
    _mock_context.get_pagerduty_webhook_url = "https://events.pagerduty.com/v2/enqueue"
    _mock_context.get_pagerduty_integration_key = "test_integration_key"

    # Test with job start message
    _config_args = {"message": "Spark expectations job has started \n\ntable_name: test_table\nrun_id: 123"}

    with patch.object(requests, "post") as mock_post:
        pagerduty_handler.send_notification(_context=_mock_context, _config_args=_config_args)
        # Should not call post since this is not a failure
        mock_post.assert_not_called()


@patch("spark_expectations.notifications.plugins.pagerduty.SparkExpectationsContext", autospec=True, spec_set=True)
def test_send_incident_skipped_for_completion_message(_mock_context):
    """Test that PagerDuty notifications are skipped for job completion messages"""
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    _mock_context.get_enable_pagerduty = True
    _mock_context.get_pagerduty_webhook_url = "https://events.pagerduty.com/v2/enqueue"
    _mock_context.get_pagerduty_integration_key = "test_integration_key"

    # Test with job completion message
    _config_args = {"message": "Spark expectations job has been completed \n\nproduct_id: test_product"}

    with patch.object(requests, "post") as mock_post:
        pagerduty_handler.send_notification(_context=_mock_context, _config_args=_config_args)
        # Should not call post since this is not a failure
        mock_post.assert_not_called()


@patch("spark_expectations.notifications.plugins.pagerduty.SparkExpectationsContext", autospec=True)
def test_send_incident_for_job_failure_message(_mock_context):
    """Test that PagerDuty notifications are sent for job failure messages"""
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    _mock_context.get_enable_pagerduty = True
    _mock_context.get_pagerduty_webhook_url = "https://events.pagerduty.com/v2/enqueue"
    _mock_context.get_pagerduty_integration_key = "test_integration_key"
    _mock_context.product_id = "test_product"
    _mock_context.get_table_name = "test_table"

    # Test with job failure message
    _config_args = {"message": "Spark expectations job has been failed \n\nproduct_id: test_product"}

    with patch.object(requests, "post") as mock_post:
        mock_response = Mock()
        mock_response.status_code = 202
        mock_post.return_value = mock_response

        pagerduty_handler.send_notification(_context=_mock_context, _config_args=_config_args)
        # Should call post since this is a failure
        mock_post.assert_called_once()


@patch("spark_expectations.notifications.plugins.pagerduty.SparkExpectationsContext", autospec=True)
def test_send_incident_for_error_threshold_breach(_mock_context):
    """Test that PagerDuty notifications are sent for error threshold breach messages"""
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    _mock_context.get_enable_pagerduty = True
    _mock_context.get_pagerduty_webhook_url = "https://events.pagerduty.com/v2/enqueue"
    _mock_context.get_pagerduty_integration_key = "test_integration_key"
    _mock_context.product_id = "test_product"
    _mock_context.get_table_name = "test_table"

    # Test with error threshold breach message
    _config_args = {"message": "Spark expectations - dropped error percentage has been exceeded above the threshold value(5%) for `row_data` quality validation"}

    with patch.object(requests, "post") as mock_post:
        mock_response = Mock()
        mock_response.status_code = 202
        mock_post.return_value = mock_response

        pagerduty_handler.send_notification(_context=_mock_context, _config_args=_config_args)
        # Should call post since this is an error condition
        mock_post.assert_called_once()


@patch("spark_expectations.notifications.plugins.pagerduty.SparkExpectationsContext", autospec=True, spec_set=True)
def test_send_incident_skipped_for_ignored_rules_failure(_mock_context):
    """Test that PagerDuty notifications are NOT sent for ignored rules failure messages.
    
    These are informational notifications only and should not create PagerDuty incidents.
    PagerDuty incidents should only be created for actual critical failures.
    """
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    _mock_context.get_enable_pagerduty = True
    _mock_context.get_pagerduty_webhook_url = "https://events.pagerduty.com/v2/enqueue"
    _mock_context.get_pagerduty_integration_key = "test_integration_key"

    # Test with ignored rules failure message
    _config_args = {"message": "Spark expectations notification on rules which action_if_failed are set to ignore"}

    with patch.object(requests, "post") as mock_post:
        pagerduty_handler.send_notification(_context=_mock_context, _config_args=_config_args)
        # Should NOT call post since this is just an informational notification about ignored rules
        mock_post.assert_not_called()


def test_is_failure_notification():
    """Test the _is_failure_notification method with various message types"""
    pagerduty_handler = SparkExpectationsPagerDutyPluginImpl()
    
    # Test failure messages that should return True
    failure_messages = [
        "Spark expectations job has been failed",
        "Job FAILED due to error",
        "Spark expectations - dropped error percentage has been exceeded above the threshold",
        "Something went wrong and FAILED",
    ]
    
    for message in failure_messages:
        assert pagerduty_handler._is_failure_notification(message) is True, f"Should detect failure in: {message}"
    
    # Test non-failure messages that should return False
    # NOTE: Ignored rules notifications are informational only and should NOT create PagerDuty incidents
    non_failure_messages = [
        "Spark expectations job has started",
        "Spark expectations job has been completed",
        "Spark expectations notification on rules which action_if_failed are set to ignore",  # Informational only
        "Everything is working fine",
        "Job succeeded",
        "Processing data successfully",
        "",
        None,
    ]
    
    for message in non_failure_messages:
        assert pagerduty_handler._is_failure_notification(message) is False, f"Should not detect failure in: {message}"



import ast
import os
from typing import Any, Dict, Union

import yaml
from pyspark.sql.session import SparkSession

current_dir = os.path.dirname(os.path.abspath(__file__))


def load_configurations(spark: SparkSession) -> tuple[Dict[str, Union[str, int, bool, Dict[str, str]]], Dict[str, Union[str, int, bool, Dict[str, str]]]]:
    """
    Load Spark configuration settings from a YAML file and apply them to the provided SparkSession.

    This function:
    - Reads the configuration file located at `../config/spark-expectations-default-config.yaml`.
    - Separates streaming (`se.streaming.*`) and notification (`spark.expectations.*`) configurations into dictionaries.
    - Sets other configuration values directly in the Spark session.
    - Stores streaming and notification configs as JSON strings in Spark session configs.
    - In serverless environments, skips configuration setting to avoid CONFIG_NOT_AVAILABLE errors.
    - Raises RuntimeError for file not found, YAML parsing errors, permission issues, or other exceptions.

    Args:
        spark (SparkSession): The SparkSession to apply configurations to.

    Raises:
        RuntimeError: If the configuration file is not found, cannot be parsed, or other errors occur.
    """
    try:
        with open(f"{current_dir}/../config/spark-expectations-default-config.yaml", "r", encoding="utf-8") as cfg_file:
            config = yaml.safe_load(cfg_file)
        if config is None:
            config = {}
        elif not isinstance(config, dict):
            raise yaml.YAMLError("Spark config YAML file is not valid.")
        streaming_config = {}
        notification_config = {}
        for key, value in config.items():
            if key.startswith("se.streaming."):
                streaming_config[key] = value
            elif key.startswith("spark.expectations."):
                notification_config[key] = value
            else:
                spark.conf.set(key, str(value))
        
        return streaming_config, notification_config      
        

    except FileNotFoundError as e:
        raise RuntimeError(f"Spark config YAML file not found: {e}") from e
    except yaml.YAMLError as e:
        raise RuntimeError(f"Error parsing Spark config YAML configuration file: {e}") from e
    except Exception as e:
        raise RuntimeError(f"An unexpected error occurred while loading spark configurations: {e}") from e


def get_config_dict(
    spark: SparkSession, user_conf: Dict[str, Union[str, int, bool, Dict[str, str]]] = None
) -> tuple[Dict[str, Union[str, int, bool, Dict[str, str]]], Dict[str, Union[str, int, bool, Dict[str, str]]],]:
    """
    Retrieve both notification and streaming config dictionaries from the user configuration or Spark session or default configuration.

    Args:
        spark (SparkSession): The Spark session to retrieve the configuration from.
        user_conf ([Dict[str, Any]]): User configuration to merge with default configuration.

    Returns:
        tuple: A tuple containing (notification_dict, streaming_dict).

    Raises:
        RuntimeError: If there are errors parsing or retrieving the configuration.
    """

    def _build_config_dict(
        default_dict: Dict[str, Union[str, int, bool, Dict[str, str]]],
        user_conf: Dict[str, Union[str, int, bool, Dict[str, str]]] = None,
    ) -> Dict[str, Union[str, int, bool, Dict[str, str]]]:
        """Helper function to build configuration dictionary with type inference."""
        is_serverless_mode = user_conf.get("spark.expectations.is.serverless", False) if user_conf else False
        
        if user_conf:
            if is_serverless_mode:
                config_dict = {
                    key: infer_safe_cast(user_conf.get(key, str(value)))
                    for key, value in default_dict.items()
                }
            else:
                config_dict = {
                    key: infer_safe_cast(user_conf.get(key, spark.conf.get(key, str(value))))
                    for key, value in default_dict.items()
                }
        else:
            config_dict = {key: infer_safe_cast(spark.conf.get(key, str(value))) for key, value in default_dict.items()}
        return config_dict

    try:
        # Get configurations directly from load_configurations
        default_streaming_dict, default_notification_dict = load_configurations(spark)

        # Build both dictionaries using the helper function
        notification_dict = _build_config_dict(default_notification_dict, user_conf)
        streaming_dict = _build_config_dict(default_streaming_dict, user_conf)

        return notification_dict, streaming_dict

    except Exception as e:
        raise RuntimeError(f"Error retrieving configuration: {e}") from e


def get_spark_session() -> SparkSession:
    if (os.environ.get("UNIT_TESTING_ENV") == "spark_expectations_unit_testing_on_github_actions") or (
        os.environ.get("SPARKEXPECTATIONS_ENV") == "local"
    ):
        builder = (
            SparkSession.builder.config("spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension")
            .config("spark.jars.packages", "io.delta:delta-spark_2.13:4.0.0")
            .config(
                "spark.sql.catalog.spark_catalog",
                "org.apache.spark.sql.delta.catalog.DeltaCatalog",
            )
            .config("spark.sql.warehouse.dir", "/tmp/hive/warehouse")
            .config("spark.driver.extraJavaOptions", "-Dderby.system.home=/tmp/derby")
            .config("spark.jars.ivy", "/tmp/ivy2")
            .config(  # below jars are used only in the local env, not coupled with databricks or EMR
                "spark.jars",
                f"{current_dir}/../../jars/spark-sql-kafka-0-10_2.13-4.0.0.jar,"
                f"{current_dir}/../../jars/kafka-clients-3.7.0.jar,"
                f"{current_dir}/../../jars/commons-pool2-2.12.0.jar,"
                f"{current_dir}/../../jars/spark-token-provider-kafka-0-10_2.13-4.0.0.jar",
            )
            .config("spark.sql.shuffle.partitions", 1)
            .config("spark.dynamicAllocation.enabled", "false")
            .config("spark.ui.enabled", "false")
            .config("spark.ui.showConsoleProgress", "false")
        )

        sparksession = builder.getOrCreate()
    else:
        sparksession = SparkSession.getActiveSession()
    load_configurations(sparksession)  # Load default configs, return values not needed here
    return sparksession


def infer_safe_cast(input_value: Any) -> Union[int, float, bool, dict, str, None]:  # pylint: disable=R0911
    """
    Infers and safely casts the input value to int, float, bool, dict, str, or None.

    Args:
        input_value: The value to analyze (can be any type)

    Returns:
        Union[int, float, bool, dict, str, None]: The inferred and converted value
    """
    if input_value is None:
        return None

    # Return early for already acceptable types
    if isinstance(input_value, (int, float, bool, dict, list)):
        return input_value

    # Convert to string and clean
    cleaned_input = str(input_value).strip()

    # Handle string representations of None
    if cleaned_input.lower() in {"none", "null"}:
        return None

    # Handle booleans (case-insensitive)
    if cleaned_input.lower() in {"true", "false"}:
        return cleaned_input.lower() == "true"

    # Try integer
    try:
        return int(cleaned_input)
    except ValueError:
        pass

    # Try float
    try:
        return float(cleaned_input)
    except ValueError:
        pass

    # Try dictionary
    try:
        parsed = ast.literal_eval(cleaned_input)
        if isinstance(parsed, dict):
            return parsed
    except (ValueError, SyntaxError):
        pass

    # Fallback to original string (not lowercased)
    return str(input_value).strip()


import os
from unittest.mock import patch
import pytest
from pyspark.sql.functions import col
from spark_expectations.core import get_spark_session
from spark_expectations.sinks.utils.collect_statistics import (
    SparkExpectationsCollectStatistics,
)
from spark_expectations.sinks.utils.writer import SparkExpectationsWriter
from spark_expectations.core.exceptions import SparkExpectationsMiscException
from unittest.mock import Mock
from spark_expectations.core.context import SparkExpectationsContext
from spark_expectations.core.expectations import WrappedDataFrameWriter

spark = get_spark_session()


@pytest.fixture(name="_fixture_local_kafka_topic",scope="session",autouse=True)
def fixture_setup_local_kafka_topic():
    current_dir = os.path.dirname(os.path.abspath(__file__))

    if os.getenv("UNIT_TESTING_ENV") != "spark_expectations_unit_testing_on_github_actions":
        # remove if docker conatiner is running
        os.system(f"sh {current_dir}/../../../../containers/kafka/scripts/docker_kafka_stop_script.sh")

        # start docker container and create the topic
        os.system(f"sh {current_dir}/../../../../containers/kafka/scripts/docker_kafka_start_script.sh")

        yield "docker container started"

        # remove docker container
        os.system(f"sh {current_dir}/../../../../containers/kafka/scripts/docker_kafka_stop_script.sh")

    else:
        yield (
            "A Kafka server has been launched within a Docker container for the purpose of conducting tests in"
            " a Jenkins environment"
        )


@pytest.fixture(name="_fixture_create_stats_table")
def fixture_create_stats_table():
    # drop if exist dq_spark database and create with test_dq_stats_table
    os.system("rm -rf /tmp/hive/warehouse/dq_spark.db")
    spark.sql("create database if not exists dq_spark")
    spark.sql("use dq_spark")

    spark.sql("drop table if exists test_dq_stats_table")
    os.system("rm -rf /tmp/hive/warehouse/dq_spark.db/test_dq_stats_table")
    spark.sql(
        """
    create table test_dq_stats_table (
    product_id STRING,
    table_name STRING,
    input_count LONG,
    error_count LONG,
    output_count LONG,
    output_percentage FLOAT,
    success_percentage FLOAT,
    error_percentage FLOAT,
    source_agg_dq_results array<map<string, string>>,
    final_agg_dq_results array<map<string, string>>,
    source_query_dq_results array<map<string, string>>,
    final_query_dq_results array<map<string, string>>,
    row_dq_res_summary array<map<string, string>>,
    row_dq_error_threshold array<map<string, string>>,
    dq_status map<string, string>,
    dq_run_time map<string, float>,
    dq_rules map<string, map<string,int>>,
    meta_dq_run_id STRING,
    meta_dq_run_date DATE,
    meta_dq_run_datetime TIMESTAMP,
    dq_env STRING
    )
    USING delta
    """
    )

    yield "test_dq_stats_table"

    spark.sql("drop table if exists test_dq_stats_table")
    os.system("rm -rf /tmp/hive/warehouse/dq_spark.db/test_dq_stats_table")

    # remove database
    os.system("rm -rf /tmp/hive/warehouse/dq_spark.db")


@pytest.mark.parametrize(
    "input_record, expected_result",
    [
        (
            {
                "input_count": 100,
                "error_count": 10,
                "output_count": 90,
                "source_agg_results": [
                    {
                        "rule_name": "rule1",
                        "action_if_failed": "ignore",
                        "description": "not null values in col1",
                        "rule_type": "agg_dq",
                    }
                ],
                "final_agg_results": [
                    {
                        "rule_name": "rule1",
                        "action_if_failed": "ignore",
                        "description": "not null values in col1",
                        "rule_type": "agg_dq",
                    }
                ],
                "source_query_dq_results": [
                    {
                        "rule_name": "rule5",
                        "action_if_failed": "ignore",
                        "description": "sum of col5 must be gt 100",
                        "rule_type": "query_dq",
                    }
                ],
                "final_query_dq_results": [
                    {
                        "rule_name": "rule6",
                        "action_if_failed": "ignore",
                        "description": "distinct in col6 must be lt 3",
                        "rule_type": "query_dq",
                    }
                ],
                "row_dq_res_summary": [
                    {
                        "rule_name": "rule1",
                        "action_if_failed": "ignore",
                        "rule_type": "row_dq",
                        "failed_count": 10,
                    },
                    {
                        "rule_name": "rule2",
                        "action_if_failed": "drop",
                        "rule_type": "row_dq",
                        "failed_count": 5,
                    },
                    {
                        "rule_name": "rule3",
                        "action_if_failed": "ignore",
                        "rule_type": "row_dq",
                        "failed_count": 3,
                    },
                ],
                "row_dq_error_threshold": [
                    {
                        "rule_name": "rule1",
                        "action_if_failed": "ignore",
                        "description": "description1",
                        "rule_type": "row_dq",
                        "error_drop_threshold": "15",
                        "error_drop_percentage": "1.0",
                    },
                    {
                        "rule_name": "rule2",
                        "action_if_failed": "drop",
                        "description": "description2",
                        "rule_type": "row_dq",
                        "error_drop_threshold": "10",
                        "error_drop_percentage": "0.5",
                    },
                    {
                        "rule_name": "rule3",
                        "action_if_failed": "ignore",
                        "description": "description3",
                        "rule_type": "row_dq",
                        "error_drop_threshold": "5",
                        "error_drop_percentage": "0.3",
                    },
                ],
                "dq_run_time": {
                    "final_query_dq_run_time": 22.7,
                    "source_agg_dq_run_time": 17.2,
                    "row_dq_run_time": 29.3,
                    "source_query_dq_run_time": 22.4,
                    "final_agg_dq_run_time": 11.0,
                    "run_time": 108.5,
                },
                "dq_rules": {
                    "rules": {"num_dq_rules": 17, "num_row_dq_rules": 5},
                    "query_dq_rules": {
                        "num_final_query_dq_rules": 8,
                        "num_source_query_dq_rules": 3,
                        "num_query_dq_rules": 5,
                    },
                    "agg_dq_rules": {
                        "num_source_agg_dq_rules": 4,
                        "num_agg_dq_rules": 4,
                        "num_final_agg_dq_rules": 1,
                    },
                },
                "status": {
                    "run_status": "Passed",
                    "source_agg_dq": "Passed",
                    "row_dq": "Passed",
                    "final_agg_dq": "Passed",
                    "source_query_dq": "Passed",
                    "final_query_dq": "Passed",
                },
            },
            {
                "output_percentage": 90.0,
                "success_percentage": 90.0,
                "error_percentage": 10.0,
            },
        ),
        (
            {
                "input_count": 100,
                "error_count": 10,
                "output_count": 95,
                "source_agg_results": None,
                "final_agg_results": [
                    {
                        "rule_name": "rule2",
                        "action_if_failed": "drop",
                        "description": "not null values in col2",
                        "rule_type": "agg_dq",
                    }
                ],
                "source_query_dq_results": None,
                "final_query_dq_results": None,
                "row_dq_res_summary": [
                    {
                        "rule_name": "rule1",
                        "action_if_failed": "ignore",
                        "rule_type": "row_dq",
                        "failed_count": 10,
                    },
                    {
                        "rule_name": "rule2",
                        "action_if_failed": "drop",
                        "rule_type": "row_dq",
                        "failed_count": 7,
                    },
                    {
                        "rule_name": "rule3",
                        "action_if_failed": "ignore",
                        "rule_type": "row_dq",
                        "failed_count": 8,
                    },
                ],
                "row_dq_error_threshold": [
                    {
                        "rule_name": "rule1",
                        "action_if_failed": "ignore",
                        "description": "description1",
                        "rule_type": "row_dq",
                        "error_drop_threshold": "15",
                        "error_drop_percentage": "1.0",
                    },
                    {
                        "rule_name": "rule2",
                        "action_if_failed": "drop",
                        "description": "description2",
                        "rule_type": "row_dq",
                        "error_drop_threshold": "10",
                        "error_drop_percentage": "0.7",
                    },
                    {
                        "rule_name": "rule3",
                        "action_if_failed": "ignore",
                        "description": "description3",
                        "rule_type": "row_dq",
                        "error_drop_threshold": "5",
                        "error_drop_percentage": "0.8",
                    },
                ],
                "dq_run_time": {
                    "final_query_dq_run_time": 0.0,
                    "source_agg_dq_run_time": 0.0,
                    "row_dq_run_time": 29.3,
                    "source_query_dq_run_time": 0.0,
                    "final_agg_dq_run_time": 11.0,
                    "run_time": 108.5,
                },
                "dq_rules": {
                    "rules": {"num_dq_rules": 14, "num_row_dq_rules": 3},
                    "query_dq_rules": {
                        "num_final_query_dq_rules": 5,
                        "num_source_query_dq_rules": 5,
                        "num_query_dq_rules": 1,
                    },
                    "agg_dq_rules": {
                        "num_source_agg_dq_rules": 4,
                        "num_agg_dq_rules": 2,
                        "num_final_agg_dq_rules": 4,
                    },
                },
                "status": {
                    "run_status": "Passed",
                    "source_agg_dq": "Skipped",
                    "row_dq": "Passed",
                    "final_agg_dq": "Passed",
                    "source_query_dq": "Skipped",
                    "final_query_dq": "Skipped",
                },
            },
            {
                "output_percentage": 95.0,
                "success_percentage": 90.0,
                "error_percentage": 10.0,
            },
        ),
        (
            {
                "input_count": 100,
                "error_count": 100,
                "output_count": 100,
                "source_agg_results": [
                    {
                        "rule_name": "rule2",
                        "action_if_failed": "drop",
                        "description": "not null values in col2",
                        "rule_type": "agg_dq",
                    }
                ],
                "final_agg_results": None,
                "source_query_dq_results": None,
                "final_query_dq_results": [
                    {
                        "rule_name": "rule5",
                        "action_if_failed": "ignore",
                        "description": "sum of col5 must be gt 100",
                        "rule_type": "query_dq",
                    }
                ],
                "row_dq_res_summary": [
                    {
                        "rule_name": "rule1",
                        "action_if_failed": "ignore",
                        "rule_type": "row_dq",
                        "failed_count": 10,
                    },
                    {
                        "rule_name": "rule2",
                        "action_if_failed": "drop",
                        "rule_type": "row_dq",
                        "failed_count": 7,
                    },
                    {
                        "rule_name": "rule3",
                        "action_if_failed": "ignore",
                        "rule_type": "row_dq",
                        "failed_count": 8,
                    },
                ],
                "row_dq_error_threshold": [
                    {
                        "rule_name": "rule1",
                        "action_if_failed": "ignore",
                        "description": "description1",
                        "rule_type": "row_dq",
                        "error_drop_threshold": "15",
                        "error_drop_percentage": "1.0",
                    },
                    {
                        "rule_name": "rule2",
                        "action_if_failed": "drop",
                        "description": "description2",
                        "rule_type": "row_dq",
                        "error_drop_threshold": "10",
                        "error_drop_percentage": "0.7",
                    },
                    {
                        "rule_name": "rule3",
                        "action_if_failed": "ignore",
                        "description": "description3",
                        "rule_type": "row_dq",
                        "error_drop_threshold": "5",
                        "error_drop_percentage": "0.8",
                    },
                ],
                "dq_run_time": {
                    "final_query_dq_run_time": 22.7,
                    "source_agg_dq_run_time": 17.2,
                    "row_dq_run_time": 29.3,
                    "source_query_dq_run_time": 0.0,
                    "final_agg_dq_run_time": 0.0,
                    "run_time": 108.5,
                },
                "dq_rules": {
                    "rules": {"num_dq_rules": 17, "num_row_dq_rules": 10},
                    "query_dq_rules": {
                        "num_final_query_dq_rules": 5,
                        "num_source_query_dq_rules": 2,
                        "num_query_dq_rules": 3,
                    },
                    "agg_dq_rules": {
                        "num_source_agg_dq_rules": 4,
                        "num_agg_dq_rules": 2,
                        "num_final_agg_dq_rules": 2,
                    },
                },
                "status": {
                    "run_status": "Passed",
                    "source_agg_dq": "Passed",
                    "row_dq": "Passed",
                    "final_agg_dq": "Skipped",
                    "source_query_dq": "Skipped",
                    "final_query_dq": "Passed",
                },
            },
            {
                "output_percentage": 100.0,
                "success_percentage": 0.0,
                "error_percentage": 100.0,
            },
        ),
    ],
)
def test_collect_stats_on_success_failure(
    input_record,
    expected_result,
    _fixture_local_kafka_topic,
    _fixture_create_stats_table,
):
    # create mock _context object
    _mock_context = Mock(spec=SparkExpectationsContext)
    _mock_context.get_dq_rules_params = {"env": "test_env"}
    setattr(_mock_context, "get_dq_stats_table_name", "test_dq_stats_table")
    setattr(_mock_context, "get_run_date_name", "meta_dq_run_date")
    setattr(_mock_context, "get_run_date_time_name", "meta_dq_run_datetime")
    setattr(_mock_context, "get_run_date", "2022-12-27 10:39:44")
    setattr(_mock_context, "get_run_id_name", "meta_dq_run_id")
    setattr(_mock_context, "get_run_id", "product1_run_test")
    setattr(_mock_context, "get_dq_run_status", input_record.get("status").get("run_status"))
    setattr(
        _mock_context,
        "get_source_agg_dq_status",
        input_record.get("status").get("source_agg_dq"),
    )
    setattr(_mock_context, "get_row_dq_status", input_record.get("status").get("row_dq"))
    setattr(
        _mock_context,
        "get_final_agg_dq_status",
        input_record.get("status").get("final_agg_dq"),
    )
    setattr(
        _mock_context,
        "get_source_query_dq_status",
        input_record.get("status").get("source_query_dq"),
    )
    setattr(
        _mock_context,
        "get_final_query_dq_status",
        input_record.get("status").get("final_query_dq"),
    )
    setattr(_mock_context, "get_input_count", input_record.get("input_count"))
    setattr(_mock_context, "get_error_count", input_record.get("error_count"))
    setattr(_mock_context, "get_output_count", input_record.get("output_count"))
    setattr(
        _mock_context,
        "get_source_agg_dq_result",
        input_record.get("source_agg_results"),
    )
    setattr(_mock_context, "get_final_agg_dq_result", input_record.get("final_agg_results"))
    setattr(_mock_context, "get_table_name", "employee_table")
    setattr(
        _mock_context,
        "get_output_percentage",
        round(
            (input_record.get("output_count") / input_record.get("input_count")) * 100,
            2,
        ),
    )
    setattr(
        _mock_context,
        "get_error_percentage",
        round((input_record.get("error_count") / input_record.get("input_count")) * 100, 2),
    )
    setattr(
        _mock_context,
        "get_success_percentage",
        round(
            ((input_record.get("input_count") - input_record.get("error_count")) / input_record.get("input_count"))
            * 100,
            2,
        ),
    )
    setattr(_mock_context, "get_env", "local")
    setattr(
        _mock_context,
        "get_source_query_dq_result",
        input_record.get("source_query_dq_results"),
    )
    setattr(
        _mock_context,
        "get_final_query_dq_result",
        input_record.get("final_query_dq_results"),
    )
    setattr(
        _mock_context,
        "get_summarized_row_dq_res",
        input_record.get("row_dq_res_summary"),
    )
    setattr(
        _mock_context,
        "get_rules_exceeds_threshold",
        input_record.get("row_dq_error_threshold"),
    )

    setattr(
        _mock_context,
        "get_dq_run_time",
        round(input_record.get("dq_run_time").get("run_time"), 1),
    )
    setattr(
        _mock_context,
        "get_source_agg_dq_run_time",
        round(input_record.get("dq_run_time").get("source_agg_dq_run_time"), 1),
    )
    setattr(
        _mock_context,
        "get_source_query_dq_run_time",
        round(input_record.get("dq_run_time").get("source_query_dq_run_time"), 1),
    )
    setattr(
        _mock_context,
        "get_row_dq_run_time",
        round(input_record.get("dq_run_time").get("row_dq_run_time"), 1),
    )
    setattr(
        _mock_context,
        "get_final_agg_dq_run_time",
        round(input_record.get("dq_run_time").get("final_agg_dq_run_time"), 1),
    )
    setattr(
        _mock_context,
        "get_final_query_dq_run_time",
        round(input_record.get("dq_run_time").get("final_query_dq_run_time"), 1),
    )

    setattr(
        _mock_context,
        "get_num_row_dq_rules",
        input_record.get("dq_rules").get("rules").get("num_row_dq_rules"),
    )
    setattr(
        _mock_context,
        "get_num_dq_rules",
        input_record.get("dq_rules").get("rules").get("num_dq_rules"),
    )
    setattr(
        _mock_context,
        "get_num_agg_dq_rules",
        input_record.get("dq_rules").get("agg_dq_rules"),
    )
    setattr(
        _mock_context,
        "get_num_query_dq_rules",
        input_record.get("dq_rules").get("query_dq_rules"),
    )
    setattr(
        _mock_context,
        "_stats_table_writer_config",
        WrappedDataFrameWriter().mode("overwrite").format("delta").build(),
    )
    setattr(
        _mock_context,
        "get_stats_table_writer_config",
        WrappedDataFrameWriter().mode("overwrite").format("delta").build(),
    )
    setattr(_mock_context, "get_se_streaming_stats_dict", {"se.streaming.enable": True})
    setattr(_mock_context, "get_topic_name", "dq-sparkexpectations-stats")

    _mock_context.spark = spark
    _mock_context.product_id = "product1"

    writer = SparkExpectationsWriter(_mock_context)
    statistics_writer_obj = SparkExpectationsCollectStatistics(_mock_context, writer)

    @statistics_writer_obj.collect_stats_on_success_failure()
    def exec_func():
        return "success"

    result = exec_func()
    assert result == "success"

    stats_table = spark.table("test_dq_stats_table")
    assert stats_table.count() == 1
    row = stats_table.first()
    assert row.product_id == "product1"
    assert row.table_name == "employee_table"
    assert row.input_count == input_record.get("input_count")
    assert row.error_count == input_record.get("error_count")
    assert row.output_count == input_record.get("output_count")
    assert row.output_percentage == expected_result.get("output_percentage")
    assert row.success_percentage == expected_result.get("success_percentage")
    assert row.error_percentage == expected_result.get("error_percentage")
    assert row.source_agg_dq_results == input_record.get("source_agg_results")
    assert row.final_agg_dq_results == input_record.get("final_agg_results")
    assert row.source_query_dq_results == input_record.get("source_query_dq_results")
    assert row.final_query_dq_results == input_record.get("final_query_dq_results")
    assert row.dq_rules == input_record.get("dq_rules")
    # assert row.dq_run_time == input_record.get("dq_run_time")
    assert row.dq_status == input_record.get("status")
    assert row.meta_dq_run_id == "product1_run_test"

    assert (
        spark.read.format("kafka")
        .option("kafka.bootstrap.servers", "localhost:9092")
        .option("subscribe", "dq-sparkexpectations-stats")
        .option("startingOffsets", "earliest")
        .option("endingOffsets", "latest")
        .load()
        .orderBy(col("timestamp").desc())
        .limit(1)
        .selectExpr("cast(value as string) as value")
        .collect()
        == stats_table.selectExpr("to_json(struct(*)) AS value").collect()
    )


@pytest.mark.parametrize(
    "input_record, expected_result",
    [
        (
            {
                "input_count": 100,
                "error_count": 100,
                "output_count": 0,
                "source_agg_results": [
                    {
                        "rule_name": "rule2",
                        "action_if_failed": "drop",
                        "description": "not null values in col2",
                        "rule_type": "agg_dq",
                    }
                ],
                "final_agg_results": None,
                "source_query_dq_results": None,
                "final_query_dq_results": None,
                "row_dq_res_summary": [
                    {
                        "rule_name": "rule1",
                        "action_if_failed": "fail",
                        "rule_type": "row_dq",
                        "failed_count": 10,
                    },
                    {
                        "rule_name": "rule2",
                        "action_if_failed": "drop",
                        "rule_type": "row_dq",
                        "failed_count": 7,
                    },
                    {
                        "rule_name": "rule3",
                        "action_if_failed": "ignore",
                        "rule_type": "row_dq",
                        "failed_count": 8,
                    },
                ],
                "row_dq_error_threshold": [
                    {
                        "rule_name": "rule1",
                        "action_if_failed": "fail",
                        "description": "description1",
                        "rule_type": "row_dq",
                        "error_drop_threshold": "15",
                        "error_drop_percentage": "1.0",
                    },
                    {
                        "rule_name": "rule2",
                        "action_if_failed": "drop",
                        "description": "description2",
                        "rule_type": "row_dq",
                        "error_drop_threshold": "10",
                        "error_drop_percentage": "0.7",
                    },
                    {
                        "rule_name": "rule3",
                        "action_if_failed": "ignore",
                        "description": "description3",
                        "rule_type": "row_dq",
                        "error_drop_threshold": "5",
                        "error_drop_percentage": "0.8",
                    },
                ],
                "dq_run_time": {
                    "final_query_dq_run_time": 0.7,
                    "source_agg_dq_run_time": 17.2,
                    "row_dq_run_time": 29.3,
                    "source_query_dq_run_time": 0.0,
                    "final_agg_dq_run_time": 0.0,
                    "run_time": 118.5,
                },
                "dq_rules": {
                    "rules": {"num_dq_rules": 18, "num_row_dq_rules": 5},
                    "query_dq_rules": {
                        "num_final_query_dq_rules": 5,
                        "num_source_query_dq_rules": 5,
                        "num_query_dq_rules": 5,
                    },
                    "agg_dq_rules": {
                        "num_source_agg_dq_rules": 8,
                        "num_agg_dq_rules": 4,
                        "num_final_agg_dq_rules": 8,
                    },
                },
                "status": {
                    "run_status": "Failed",
                    "source_agg_dq": "Passed",
                    "row_dq": "Passed",
                    "final_agg_dq": "Skipped",
                    "source_query_dq": "Skipped",
                    "final_query_dq": "Skipped",
                },
            },
            {
                "output_percentage": 0.0,
                "success_percentage": 0.0,
                "error_percentage": 100.0,
            },
        ),
        (
            {
                "input_count": 100,
                "error_count": 100,
                "output_count": 0,
                "source_agg_results": [
                    {
                        "rule_name": "rule2",
                        "action_if_failed": "drop",
                        "description": "not null values in col2",
                        "rule_type": "agg_dq",
                    }
                ],
                "final_agg_results": None,
                "source_query_dq_results": [
                    {
                        "rule_name": "rule5",
                        "action_if_failed": "ignore",
                        "description": "sum of col5 must be gt 100",
                        "rule_type": "query_dq",
                    }
                ],
                "final_query_dq_results": [
                    {
                        "rule_name": "rule5",
                        "action_if_failed": "ignore",
                        "description": "sum of col5 must be gt 100",
                        "rule_type": "query_dq",
                    }
                ],
                "row_dq_res_summary": [
                    {
                        "rule_name": "rule1",
                        "action_if_failed": "ignore",
                        "rule_type": "row_dq",
                        "failed_count": 100,
                    },
                    {
                        "rule_name": "rule2",
                        "action_if_failed": "drop",
                        "rule_type": "row_dq",
                        "failed_count": 100,
                    },
                    {
                        "rule_name": "rule3",
                        "action_if_failed": "ignore",
                        "rule_type": "row_dq",
                        "failed_count": 88,
                    },
                ],
                "row_dq_error_threshold": [
                    {
                        "rule_name": "rule1",
                        "action_if_failed": "ignore",
                        "description": "description1",
                        "rule_type": "row_dq",
                        "error_drop_threshold": "15",
                        "error_drop_percentage": "100.0",
                    },
                    {
                        "rule_name": "rule2",
                        "action_if_failed": "drop",
                        "description": "description2",
                        "rule_type": "row_dq",
                        "error_drop_threshold": "10",
                        "error_drop_percentage": "100.0",
                    },
                    {
                        "rule_name": "rule3",
                        "action_if_failed": "ignore",
                        "description": "description3",
                        "rule_type": "row_dq",
                        "error_drop_threshold": "5",
                        "error_drop_percentage": "88.0",
                    },
                ],
                "dq_run_time": {
                    "final_query_dq_run_time": 22.7,
                    "source_agg_dq_run_time": 0.0,
                    "row_dq_run_time": 29.3,
                    "source_query_dq_run_time": 10.8,
                    "final_agg_dq_run_time": 0.0,
                    "run_time": 108.5,
                },
                "dq_rules": {
                    "rules": {"num_dq_rules": 18, "num_row_dq_rules": 8},
                    "query_dq_rules": {
                        "num_final_query_dq_rules": 6,
                        "num_source_query_dq_rules": 5,
                        "num_query_dq_rules": 3,
                    },
                    "agg_dq_rules": {
                        "num_source_agg_dq_rules": 4,
                        "num_agg_dq_rules": 4,
                        "num_final_agg_dq_rules": 3,
                    },
                },
                "status": {
                    "run_status": "Failed",
                    "source_agg_dq": "Passed",
                    "row_dq": "Passed",
                    "final_agg_dq": "Failed",
                    "source_query_dq": "Passed",
                    "final_query_dq": "Passed",
                },
            },
            {
                "output_percentage": 0.0,
                "success_percentage": 0.0,
                "error_percentage": 100.0,
            },
        ),
        (
            {
                "input_count": 100,
                "error_count": 100,
                "output_count": 0,
                "source_agg_results": None,
                "final_agg_results": None,
                "source_query_dq_results": None,
                "final_query_dq_results": None,
                "row_dq_res_summary": [
                    {
                        "rule_name": "rule1",
                        "action_if_failed": "ignore",
                        "rule_type": "row_dq",
                        "failed_count": 100,
                    },
                    {
                        "rule_name": "rule2",
                        "action_if_failed": "drop",
                        "rule_type": "row_dq",
                        "failed_count": 100,
                    },
                    {
                        "rule_name": "rule3",
                        "action_if_failed": "ignore",
                        "rule_type": "row_dq",
                        "failed_count": 88,
                    },
                    {
                        "rule_name": "rule4",
                        "action_if_failed": "fail",
                        "rule_type": "row_dq",
                        "failed_count": 60,
                    },
                ],
                "row_dq_error_threshold": [
                    {
                        "rule_name": "rule1",
                        "action_if_failed": "ignore",
                        "description": "description1",
                        "rule_type": "row_dq",
                        "error_drop_threshold": "15",
                        "error_drop_percentage": "100.0",
                    },
                    {
                        "rule_name": "rule2",
                        "action_if_failed": "drop",
                        "description": "description2",
                        "rule_type": "row_dq",
                        "error_drop_threshold": "10",
                        "error_drop_percentage": "100.0",
                    },
                    {
                        "rule_name": "rule3",
                        "action_if_failed": "ignore",
                        "description": "description3",
                        "rule_type": "row_dq",
                        "error_drop_threshold": "5",
                        "error_drop_percentage": "88.0",
                    },
                    {
                        "rule_name": "rule4",
                        "action_if_failed": "ignore",
                        "description": "description4",
                        "rule_type": "row_dq",
                        "error_drop_threshold": "5",
                        "error_drop_percentage": "60.0",
                    },
                ],
                "dq_run_time": {
                    "final_query_dq_run_time": 0.0,
                    "source_agg_dq_run_time": 0.0,
                    "row_dq_run_time": 29.3,
                    "source_query_dq_run_time": 0.0,
                    "final_agg_dq_run_time": 0.0,
                    "run_time": 108.5,
                },
                "dq_rules": {
                    "rules": {"num_dq_rules": 23, "num_row_dq_rules": 5},
                    "query_dq_rules": {
                        "num_final_query_dq_rules": 10,
                        "num_source_query_dq_rules": 5,
                        "num_query_dq_rules": 5,
                    },
                    "agg_dq_rules": {
                        "num_source_agg_dq_rules": 8,
                        "num_agg_dq_rules": 4,
                        "num_final_agg_dq_rules": 4,
                    },
                },
                "status": {
                    "run_status": "Failed",
                    "source_agg_dq": "Skipped",
                    "row_dq": "Failed",
                    "final_agg_dq": "Skipped",
                    "source_query_dq": "Skipped",
                    "final_query_dq": "Skipped",
                },
            },
            {
                "output_percentage": 0.0,
                "success_percentage": 0.0,
                "error_percentage": 100.0,
            },
        ),
    ],
)
def test_collect_stats_on_success_failure_exception(
    input_record,
    expected_result,
    _fixture_local_kafka_topic,
    _fixture_create_stats_table,
):
    _mock_context = Mock(spec=SparkExpectationsContext)
    _mock_context.get_dq_rules_params = {"env": "test_env"}
    setattr(_mock_context, "get_dq_stats_table_name", "test_dq_stats_table")
    setattr(_mock_context, "get_run_date_name", "meta_dq_run_date")
    setattr(_mock_context, "get_run_date_time_name", "meta_dq_run_datetime")
    setattr(_mock_context, "get_run_date", "2022-12-27 10:39:44")
    setattr(_mock_context, "get_run_id_name", "meta_dq_run_id")
    setattr(_mock_context, "get_run_id", "product1_run_test")
    setattr(_mock_context, "get_dq_run_status", input_record.get("status").get("run_status"))
    setattr(
        _mock_context,
        "get_source_agg_dq_status",
        input_record.get("status").get("source_agg_dq"),
    )
    setattr(_mock_context, "get_row_dq_status", input_record.get("status").get("row_dq"))
    setattr(
        _mock_context,
        "get_final_agg_dq_status",
        input_record.get("status").get("final_agg_dq"),
    )
    setattr(
        _mock_context,
        "get_source_query_dq_status",
        input_record.get("status").get("source_query_dq"),
    )
    setattr(
        _mock_context,
        "get_final_query_dq_status",
        input_record.get("status").get("final_query_dq"),
    )
    setattr(_mock_context, "get_input_count", input_record.get("input_count"))
    setattr(_mock_context, "get_error_count", input_record.get("error_count"))
    setattr(_mock_context, "get_output_count", input_record.get("output_count"))
    setattr(
        _mock_context,
        "get_source_agg_dq_result",
        input_record.get("source_agg_results"),
    )
    setattr(_mock_context, "get_final_agg_dq_result", input_record.get("final_agg_results"))
    setattr(_mock_context, "get_table_name", "employee_table")
    setattr(
        _mock_context,
        "get_output_percentage",
        round(
            (input_record.get("output_count") / input_record.get("input_count")) * 100,
            2,
        ),
    )
    setattr(
        _mock_context,
        "get_error_percentage",
        round((input_record.get("error_count") / input_record.get("input_count")) * 100, 2),
    )
    setattr(
        _mock_context,
        "get_success_percentage",
        round(
            ((input_record.get("input_count") - input_record.get("error_count")) / input_record.get("input_count"))
            * 100,
            2,
        ),
    )
    setattr(_mock_context, "get_env", "local")
    setattr(
        _mock_context,
        "get_source_query_dq_result",
        input_record.get("source_query_dq_results"),
    )
    setattr(
        _mock_context,
        "get_final_query_dq_result",
        input_record.get("final_query_dq_results"),
    )
    setattr(
        _mock_context,
        "get_summarized_row_dq_res",
        input_record.get("row_dq_res_summary"),
    )
    setattr(
        _mock_context,
        "get_rules_exceeds_threshold",
        input_record.get("row_dq_error_threshold"),
    )

    setattr(
        _mock_context,
        "get_dq_run_time",
        round(input_record.get("dq_run_time").get("run_time"), 1),
    )
    setattr(
        _mock_context,
        "get_source_agg_dq_run_time",
        round(input_record.get("dq_run_time").get("source_agg_dq_run_time"), 1),
    )
    setattr(
        _mock_context,
        "get_source_query_dq_run_time",
        round(input_record.get("dq_run_time").get("source_query_dq_run_time"), 1),
    )
    setattr(
        _mock_context,
        "get_row_dq_run_time",
        round(input_record.get("dq_run_time").get("row_dq_run_time"), 1),
    )
    setattr(
        _mock_context,
        "get_final_agg_dq_run_time",
        round(input_record.get("dq_run_time").get("final_agg_dq_run_time"), 1),
    )
    setattr(
        _mock_context,
        "get_final_query_dq_run_time",
        round(input_record.get("dq_run_time").get("final_query_dq_run_time"), 1),
    )

    setattr(
        _mock_context,
        "get_num_row_dq_rules",
        input_record.get("dq_rules").get("rules").get("num_row_dq_rules"),
    )
    setattr(
        _mock_context,
        "get_num_dq_rules",
        input_record.get("dq_rules").get("rules").get("num_dq_rules"),
    )
    setattr(
        _mock_context,
        "get_num_agg_dq_rules",
        input_record.get("dq_rules").get("agg_dq_rules"),
    )
    setattr(
        _mock_context,
        "get_num_query_dq_rules",
        input_record.get("dq_rules").get("query_dq_rules"),
    )
    _mock_context.spark = spark
    _mock_context.product_id = "product1"
    setattr(
        _mock_context,
        "_stats_table_writer_config",
        WrappedDataFrameWriter().mode("overwrite").format("delta").build(),
    )
    setattr(
        _mock_context,
        "get_stats_table_writer_config",
        WrappedDataFrameWriter().mode("overwrite").format("delta").build(),
    )
    setattr(_mock_context, "get_se_streaming_stats_dict", {"se.streaming.enable": True})
    setattr(_mock_context, "get_topic_name", "dq-sparkexpectations-stats")

    writer = SparkExpectationsWriter(_mock_context)
    statistics_writer_obj = SparkExpectationsCollectStatistics(_mock_context, writer)

    @statistics_writer_obj.collect_stats_on_success_failure()
    def func_exception():
        raise SparkExpectationsMiscException("test exception")

    with pytest.raises(SparkExpectationsMiscException):
        func_exception()

    stats_table = spark.table("test_dq_stats_table")
    assert stats_table.count() == 1
    row = stats_table.first()
    assert row.product_id == "product1"
    assert row.table_name == "employee_table"
    assert row.input_count == input_record.get("input_count")
    assert row.error_count == input_record.get("error_count")
    assert row.output_count == input_record.get("output_count")
    assert row.output_percentage == expected_result.get("output_percentage")
    assert row.success_percentage == expected_result.get("success_percentage")
    assert row.error_percentage == expected_result.get("error_percentage")
    assert row.source_agg_dq_results == input_record.get("source_agg_results")
    assert row.final_agg_dq_results == input_record.get("final_agg_results")
    assert row.source_query_dq_results == input_record.get("source_query_dq_results")
    assert row.final_query_dq_results == input_record.get("final_query_dq_results")
    assert row.dq_rules == input_record.get("dq_rules")
    # assert row.dq_run_time == input_record.get("dq_run_time")
    assert row.dq_status == input_record.get("status")
    assert row.meta_dq_run_id == "product1_run_test"

    assert (
        spark.read.format("kafka")
        .option("kafka.bootstrap.servers", "localhost:9092")
        .option("subscribe", "dq-sparkexpectations-stats")
        .option("startingOffsets", "earliest")
        .option("endingOffsets", "latest")
        .load()
        .orderBy(col("timestamp").desc())
        .limit(1)
        .selectExpr("cast(value as string) as value")
        .collect()
        == stats_table.selectExpr("to_json(struct(*)) AS value").collect()
    )



from unittest.mock import patch, MagicMock
import pytest
from spark_expectations.core.exceptions import SparkExpectationsEmailException
from spark_expectations.notifications.plugins.email import SparkExpectationsEmailPluginImpl


@patch("spark_expectations.notifications.plugins.email.SparkExpectationsContext", autospec=True, spec_set=True)
def test_send_notification_success(_mock_context):
    # arrange
    email_handler = SparkExpectationsEmailPluginImpl()
    _mock_context.get_enable_mail = True
    _mock_context.get_mail_from = "sender@example.com"
    _mock_context.get_to_mail = "receiver1@example.com, receiver2@example.com"
    _mock_context.get_mail_subject = "Test Email"
    _mock_context.get_mail_smtp_server = "mailhost.example.com"
    _mock_context.get_mail_smtp_port = 587

    mock_config_args = {"message": "Test Email Body"}

    with (
        patch("spark_expectations.notifications.plugins.email.smtplib.SMTP") as mock_smtp,
        patch("spark_expectations.notifications.plugins.email.MIMEMultipart") as _mock_mltp,
    ):
        # act
        email_handler.send_notification(_context=_mock_context, _config_args=mock_config_args)

        # assert
        mock_smtp.assert_called_with(_mock_context.get_mail_smtp_server, _mock_context.get_mail_smtp_port)
        mock_smtp().starttls.assert_called()
        mock_smtp().sendmail.assert_called_with(
            _mock_context.get_mail_from,
            [email.strip() for email in _mock_context.get_to_mail.split(",")],
            _mock_mltp().as_string(),
        )
        mock_smtp().quit.assert_called()


@patch("spark_expectations.notifications.plugins.email.SparkExpectationsContext", autospec=True, spec_set=True)
def test_send_notification_disable_mail(_mock_context):
    # arrange
    email_handler = SparkExpectationsEmailPluginImpl()
    _mock_context.get_enable_mail = False

    mock_config_args = {"message": "Test Email Body"}

    with patch("spark_expectations.notifications.plugins.email.smtplib.SMTP") as mock_smtp:
        # act
        email_handler.send_notification(_mock_context, mock_config_args)

        # assert
        mock_smtp.assert_not_called()


@patch("spark_expectations.notifications.plugins.email.SparkExpectationsContext", autospec=True, spec_set=True)
def test_send_notification_exception(_mock_context):
    # arrange
    email_handler = SparkExpectationsEmailPluginImpl()
    _mock_context.get_enable_mail = True
    _mock_context.get_mail_from = "sender@example.com"
    _mock_context.get_to_mail = "receiver@example.com"
    _mock_context.get_mail_subject = "Test Email"
    _mock_context.get_mail_smtp_server = "mailhost.example.com"
    _mock_context.get_mail_smtp_port = 587

    mock_config_args = {"message": "Test Email Body"}

    with (
        patch("spark_expectations.notifications.plugins.email.smtplib.SMTP") as mock_smtp,
        pytest.raises(SparkExpectationsEmailException),
    ):
        mock_smtp.side_effect = Exception("Test Exception")
        # act
        email_handler.send_notification(_mock_context, mock_config_args)

        # assert
        mock_smtp.assert_called_with(_mock_context.get_mail_smtp_server, _mock_context.get_mail_smtp_port)


@patch("spark_expectations.notifications.plugins.email.SparkExpectationsContext", autospec=True, spec_set=True)
def test_send_notification_with_smtp_auth(_mock_context):
    # arrange
    email_handler = SparkExpectationsEmailPluginImpl()
    _mock_context.get_enable_mail = True
    _mock_context.get_mail_from = "sender@example.com"
    _mock_context.get_to_mail = "receiver1@example.com, receiver2@example.com"
    _mock_context.get_mail_subject = "Test Email"
    _mock_context.get_mail_smtp_server = "mailhost.example.com"
    _mock_context.get_mail_smtp_port = 587
    _mock_context.get_enable_smtp_server_auth = True

    mock_config_args = {"message": "Test Email Body"}

    with (
        patch("spark_expectations.notifications.plugins.email.smtplib.SMTP") as mock_smtp,
        patch("spark_expectations.notifications.plugins.email.MIMEMultipart") as _mock_mltp,
        patch(
            "spark_expectations.notifications.plugins.email.SparkExpectationsEmailPluginImpl._get_smtp_password"
        ) as mock_get_smtp_password,
    ):
        # act
        email_handler.send_notification(_context=_mock_context, _config_args=mock_config_args)

        # assert
        mock_smtp.assert_called_with(_mock_context.get_mail_smtp_server, _mock_context.get_mail_smtp_port)
        mock_smtp().starttls.assert_called()
        mock_get_smtp_password.assert_called_once_with(_mock_context, mock_smtp.return_value)
        mock_smtp().sendmail.assert_called_with(
            _mock_context.get_mail_from,
            [email.strip() for email in _mock_context.get_to_mail.split(",")],
            _mock_mltp().as_string(),
        )
        mock_smtp().quit.assert_called()


@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend", autospec=True, spec_set=True)
@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend.get_secret", autospec=True, spec_set=True)
def test_get_cerberus_password(_mock_secret_handler, _mock_get_secret):
    email_handler = SparkExpectationsEmailPluginImpl()
    smtp_creds_dict = {
        "se.streaming.secret.type": "cerberus",
        "se.streaming.cerberus.url": "https://xyz.com",
        "se.streaming.cerberus.sdb.path": "app/project/env",
        "spark.expectations.notifications.cerberus.smtp.password": "password_key",
    }
    _mock_get_secret.return_value = {"password_key": "test_password"}
    _mock_secret_handler.get_secret = _mock_get_secret

    password = email_handler._get_cerberus_password(_mock_secret_handler, smtp_creds_dict)
    assert password == "test_password"


@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend", autospec=True, spec_set=True)
@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend.get_secret", autospec=True, spec_set=True)
def test_get_cerberus_password_none(_mock_secret_handler, _mock_get_secret):
    email_handler = SparkExpectationsEmailPluginImpl()
    smtp_creds_dict = {
        "se.streaming.secret.type": "cerberus",
        "se.streaming.cerberus.url": "https://xyz.com",
        "se.streaming.cerberus.sdb.path": "app/project/env",
        "spark.expectations.notifications.cerberus.smtp.password": None,
    }
    _mock_get_secret.return_value = None
    _mock_secret_handler.get_secret = _mock_get_secret

    password = email_handler._get_cerberus_password(_mock_secret_handler, smtp_creds_dict)
    assert password is None


@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend", autospec=True, spec_set=True)
@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend.get_secret", autospec=True, spec_set=True)
def test_get_databricks_password(_mock_secret_handler, _mock_get_secret):
    email_handler = SparkExpectationsEmailPluginImpl()
    smtp_creds_dict = {
        "se.streaming.secret.type": "databricks",
        "se.streaming.dbx.workspace.url": "https://xyz.databricks.com",
        "se.streaming.dbx.secret.scope": "my_secret_scope",
        "spark.expectations.notifications.dbx.smtp.password": "password_key",
    }
    _mock_get_secret.return_value = "test_password"
    _mock_secret_handler.get_secret = _mock_get_secret

    password = email_handler._get_databricks_password(_mock_secret_handler, smtp_creds_dict)
    assert password == "test_password"


@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend", autospec=True, spec_set=True)
@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend.get_secret", autospec=True, spec_set=True)
def test_get_databricks_password_none(_mock_secret_handler, _mock_get_secret):
    email_handler = SparkExpectationsEmailPluginImpl()
    smtp_creds_dict = {
        "se.streaming.secret.type": "databricks",
        "se.streaming.dbx.workspace.url": "https://xyz.databricks.com",
        "se.streaming.dbx.secret.scope": "my_secret_scope",
        # Intentionally omitting the key to cover the last line
    }
    _mock_get_secret.return_value = None
    _mock_secret_handler.get_secret = _mock_get_secret

    password = email_handler._get_databricks_password(_mock_secret_handler, smtp_creds_dict)
    assert password is None


@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend", autospec=True, spec_set=True)
def test_retrieve_password_cerberus(_mock_secret_handler):
    email_handler = SparkExpectationsEmailPluginImpl()
    secret_type = "cerberus"
    smtp_secret_dict = {
        "se.streaming.secret.type": "cerberus",
        "se.streaming.cerberus.url": "https://xyz.com",
        "se.streaming.cerberus.sdb.path": "app/project/env",
        "spark.expectations.notifications.cerberus.smtp.password": "password_key",
    }
    with patch(
        "spark_expectations.notifications.plugins.email.SparkExpectationsEmailPluginImpl._get_cerberus_password"
    ) as mock_get_cerberus_password:
        email_handler._retrieve_password(_mock_secret_handler, secret_type, smtp_secret_dict)
        mock_get_cerberus_password.assert_called_once_with(_mock_secret_handler, smtp_secret_dict)


@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend", autospec=True, spec_set=True)
def test_retrieve_password_databricks(_mock_secret_handler):
    email_handler = SparkExpectationsEmailPluginImpl()
    secret_type = "databricks"
    smtp_secret_dict = {
        "se.streaming.secret.type": "databricks",
        "se.streaming.dbx.workspace.url": "https://xyz.databricks.com",
        "se.streaming.dbx.secret.scope": "my_secret_scope",
        "spark.expectations.notifications.dbx.smtp.password": "password_key",
    }
    with patch(
        "spark_expectations.notifications.plugins.email.SparkExpectationsEmailPluginImpl._get_databricks_password"
    ) as mock_get_databricks_password:
        email_handler._retrieve_password(_mock_secret_handler, secret_type, smtp_secret_dict)
        mock_get_databricks_password.assert_called_once_with(_mock_secret_handler, smtp_secret_dict)


@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend", autospec=True, spec_set=True)
@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend.get_secret", autospec=True, spec_set=True)
def test_get_databricks_password(_mock_secret_handler, _mock_get_secret):
    email_handler = SparkExpectationsEmailPluginImpl()
    smtp_creds_dict = {
        "se.streaming.secret.type": "databricks",
        "se.streaming.dbx.workspace.url": "https://xyz.databricks.com",
        "se.streaming.dbx.secret.scope": "my_secret_scope",
        "spark.expectations.notifications.dbx.smtp.password": "password_key",
    }
    _mock_get_secret.return_value = "test_password"
    _mock_secret_handler.get_secret = _mock_get_secret

    password = email_handler._get_databricks_password(_mock_secret_handler, smtp_creds_dict)
    assert password == "test_password"


@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend", autospec=True, spec_set=True)
@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend.get_secret", autospec=True, spec_set=True)
def test_get_databricks_password_none(_mock_secret_handler, _mock_get_secret):
    email_handler = SparkExpectationsEmailPluginImpl()
    smtp_creds_dict = {
        "se.streaming.secret.type": "databricks",
        "se.streaming.dbx.workspace.url": "https://xyz.databricks.com",
        "se.streaming.dbx.secret.scope": "my_secret_scope",
        # Intentionally omitting the key to cover the last line
    }
    _mock_get_secret.return_value = None
    _mock_secret_handler.get_secret = _mock_get_secret

    password = email_handler._get_databricks_password(_mock_secret_handler, smtp_creds_dict)
    assert password is None


@patch("spark_expectations.secrets.SparkExpectationsSecretsBackend", autospec=True, spec_set=True)
def test_retrieve_password_none(_mock_secret_handler):
    email_handler = SparkExpectationsEmailPluginImpl()
    secret_type = "not_databricks"
    smtp_secret_dict = {
        "se.streaming.secret.type": "not_databricks",
        "se.streaming.dbx.workspace.url": "https://xyz.databricks.com",
        "se.streaming.dbx.secret.scope": "my_secret_scope",
        "spark.expectations.notifications.dbx.smtp.password": "password_key",
    }

    password = email_handler._retrieve_password(_mock_secret_handler, secret_type, smtp_secret_dict)
    assert password is None


@patch("spark_expectations.notifications.plugins.email.SparkExpectationsContext", autospec=True, spec_set=True)
def test_get_smtp_password(_mock_context):
    email_handler = SparkExpectationsEmailPluginImpl()
    _mock_context.get_mail_from = "test@example.com"
    _mock_context.get_mail_smtp_password = "test_password"

    with patch("spark_expectations.notifications.plugins.email.smtplib.SMTP") as mock_smtp:
        server = mock_smtp.return_value
        email_handler._get_smtp_password(_mock_context, server)
        server.login.assert_called_once_with("test@example.com", "test_password")


@patch("spark_expectations.notifications.plugins.email.SparkExpectationsContext", autospec=True, spec_set=True)
def test_get_smtp_password_with_retrieve_method(_mock_context):
    email_handler = SparkExpectationsEmailPluginImpl()
    _mock_context.get_mail_from = "test@example.com"
    _mock_context.get_mail_smtp_password = None
    _mock_context.get_smtp_creds_dict = {
        "se.streaming.secret.type": "databricks",
        "se.streaming.dbx.workspace.url": "https://xyz.databricks.com",
        "se.streaming.dbx.secret.scope": "my_secret_scope",
        "spark.expectations.notifications.dbx.smtp.password": "password_key",
    }

    with (
        patch("spark_expectations.notifications.plugins.email.smtplib.SMTP") as mock_smtp,
        patch(
            "spark_expectations.notifications.plugins.email.SparkExpectationsEmailPluginImpl._retrieve_password"
        ) as mock_retrieve_password,
    ):
        server = mock_smtp.return_value
        mock_retrieve_password.return_value = "test_password"
        email_handler._get_smtp_password(_mock_context, server)
        server.login.assert_called_once_with("test@example.com", "test_password")


@patch("spark_expectations.notifications.plugins.email.SparkExpectationsContext", autospec=True, spec_set=True)
def test_get_smtp_password_missing_key(_mock_context):
    email_handler = SparkExpectationsEmailPluginImpl()
    _mock_context.get_mail_from = "test@example.com"
    _mock_context.get_mail_smtp_password = None
    _mock_context.get_smtp_creds_dict = {
        "se.streaming.secret.type": "databricks",
        "se.streaming.dbx.workspace.url": "https://xyz.databricks.com",
        "se.streaming.dbx.secret.scope": "my_secret_scope",
        "spark.expectations.notifications.dbx.smtp.password": None,
    }

    with (
        patch("spark_expectations.notifications.plugins.email.smtplib.SMTP") as mock_smtp,
        patch(
            "spark_expectations.notifications.plugins.email.SparkExpectationsEmailPluginImpl._retrieve_password",
            side_effect=KeyError,
        ),
    ):
        server = mock_smtp.return_value

        with pytest.raises(SparkExpectationsEmailException, match="SMTP password key is missing in the secret."):
            email_handler._get_smtp_password(_mock_context, server)


@patch("spark_expectations.notifications.plugins.email.SparkExpectationsContext", autospec=True, spec_set=True)
def test_get_smtp_password_failed_retrieve(_mock_context):
    email_handler = SparkExpectationsEmailPluginImpl()
    _mock_context.get_mail_from = "test@example.com"
    _mock_context.get_mail_smtp_password = None
    _mock_context.get_smtp_creds_dict = {
        "se.streaming.secret.type": "databricks",
        "se.streaming.dbx.workspace.url": "https://xyz.databricks.com",
        "se.streaming.dbx.secret.scope": "my_secret_scope",
        "spark.expectations.notifications.dbx.smtp.password": "password_key",
    }

    with (
        patch("spark_expectations.notifications.plugins.email.smtplib.SMTP") as mock_smtp,
        patch(
            "spark_expectations.notifications.plugins.email.SparkExpectationsEmailPluginImpl._retrieve_password",
            side_effect=Exception("Generic error"),
        ),
    ):
        server = mock_smtp.return_value

        with pytest.raises(SparkExpectationsEmailException, match="Failed to retrieve SMTP password."):
            email_handler._get_smtp_password(_mock_context, server)


@patch("spark_expectations.notifications.plugins.email.SparkExpectationsContext", autospec=True, spec_set=True)
def test_get_smtp_password_none_exception(_mock_context):
    email_handler = SparkExpectationsEmailPluginImpl()
    _mock_context.get_mail_from = "test@example.com"
    _mock_context.get_mail_smtp_password = None
    _mock_context.get_smtp_creds_dict = {
        "se.streaming.secret.type": "databricks",
        "se.streaming.dbx.workspace.url": "https://xyz.databricks.com",
        "se.streaming.dbx.secret.scope": "my_secret_scope",
        "spark.expectations.notifications.dbx.smtp.password": "password_key",
    }

    with (
        patch("spark_expectations.notifications.plugins.email.smtplib.SMTP") as mock_smtp,
        patch(
            "spark_expectations.notifications.plugins.email.SparkExpectationsEmailPluginImpl._retrieve_password"
        ) as mock_retrieve_password,
    ):
        server = mock_smtp.return_value
        mock_retrieve_password.return_value = None

        with pytest.raises(SparkExpectationsEmailException, match="SMTP password is not set."):
            email_handler._get_smtp_password(_mock_context, server)


@patch("spark_expectations.notifications.plugins.email.SparkExpectationsContext", autospec=True, spec_set=True)
def test_process_message_plain_text(_mock_context):
    """Test that _process_message returns plain text content when content_type is not html"""
    email_handler = SparkExpectationsEmailPluginImpl()
    _mock_context.get_enable_templated_basic_email_body = False

    config_args = {"message": "Test message content", "content_type": "plain"}

    mail_content, content_type = email_handler._process_message(_mock_context, config_args)

    assert mail_content == "Test message content"
    assert content_type == "plain"


@patch("spark_expectations.notifications.plugins.email.SparkExpectationsContext", autospec=True, spec_set=True)
def test_process_message_html_content(_mock_context):
    """Test that _process_message keeps html content_type when specified"""
    email_handler = SparkExpectationsEmailPluginImpl()
    _mock_context.get_enable_templated_basic_email_body = False

    config_args = {"message": "<p>Test HTML content</p>", "content_type": "html"}

    mail_content, content_type = email_handler._process_message(_mock_context, config_args)

    assert mail_content == "<p>Test HTML content</p>"
    assert content_type == "html"


@patch("spark_expectations.notifications.plugins.email.SparkExpectationsContext", autospec=True, spec_set=True)
def test_process_message_detailed_notification(_mock_context):
    """Test that template is not used when email_notification_type is 'detailed'"""
    email_handler = SparkExpectationsEmailPluginImpl()
    _mock_context.get_enable_templated_basic_email_body = True
    _mock_context.get_basic_default_template = None

    config_args = {
        "message": "Test detailed message",
        "content_type": "html",
        "email_notification_type": "detailed",
    }

    mail_content, content_type = email_handler._process_message(_mock_context, config_args)

    assert mail_content == "Test detailed message"
    assert content_type == "html"


@patch("spark_expectations.notifications.plugins.email.Environment")
@patch("spark_expectations.notifications.plugins.email.PackageLoader")
@patch("spark_expectations.notifications.plugins.email.SparkExpectationsContext", autospec=True, spec_set=True)
def test_process_message_with_template_filesystem(_mock_context, mock_fs_loader, mock_env):
    """Test that template is used when email_notification_type is not 'detailed' and template enabled"""
    email_handler = SparkExpectationsEmailPluginImpl()
    _mock_context.get_enable_templated_basic_email_body = True
    _mock_context.get_basic_default_template = None
    _mock_context.get_mail_subject = "Test Subject"

    # Setup template mocks
    mock_template = MagicMock()
    mock_template.module.render_table = "render_function"
    mock_template.render.return_value = "<table>Rendered HTML</table>"
    mock_env.return_value.get_template.return_value = mock_template

    # Test message with format similar to your expected format
    test_message = "Title Line\ntable_name: test_table\nrun_id: 12345"

    config_args = {
        "message": test_message,
        "content_type": "plain",
    }

    mail_content, content_type = email_handler._process_message(_mock_context, config_args)

    # Check that the template was used
    mock_fs_loader.assert_called_once_with("spark_expectations", "config/templates")
    mock_env.assert_called_once_with(loader=mock_fs_loader.return_value)
    mock_env.return_value.get_template.assert_called_once_with("basic_email_alert_template.jinja")
    mock_template.render.assert_called_once()

    # Check the resulting content
    assert "<h2>Test Subject</h2>" in mail_content
    assert "<table>Rendered HTML</table>" in mail_content
    assert content_type == "html"

    # Verify the message data was parsed correctly
    call_args = mock_template.render.call_args[1]
    assert call_args["title"] == "Title Line"
    assert call_args["rows"] == [["table_name", " test_table"], ["run_id", " 12345"]]
    assert call_args["render_table"] == "render_function"


@patch("spark_expectations.notifications.plugins.email.Environment")
@patch("spark_expectations.notifications.plugins.email.BaseLoader")
@patch("spark_expectations.notifications.plugins.email.SparkExpectationsContext", autospec=True, spec_set=True)
def test_process_message_with_template_baseloader(_mock_context, mock_base_loader, mock_env):
    """Test template processing when a custom template is provided"""
    email_handler = SparkExpectationsEmailPluginImpl()
    _mock_context.get_enable_templated_basic_email_body = True
    _mock_context.get_basic_default_template = "Custom template string"
    _mock_context.get_mail_subject = "Custom Subject"

    # Setup template mocks
    mock_template = MagicMock()
    mock_template.module.render_table = "render_function"
    mock_template.render.return_value = "<div>Custom Template</div>"
    mock_env.return_value.from_string.return_value = mock_template

    test_message = "Alert\nkey1: value1\nkey2: value2"

    config_args = {
        "message": test_message,
        "content_type": "plain",
    }

    mail_content, content_type = email_handler._process_message(_mock_context, config_args)

    # Check the template was used with BaseLoader
    mock_env.return_value.from_string.assert_called_with("Custom template string")
    mock_template.render.assert_called_once()
    assert "<h2>Custom Subject</h2>" in mail_content
    assert "<div>Custom Template</div>" in mail_content
    assert content_type == "html"

    # Verify message parsing
    call_args = mock_template.render.call_args[1]
    assert call_args["title"] == "Alert"
    assert call_args["rows"] == [["key1", " value1"], ["key2", " value2"]]
    assert call_args["render_table"] == "render_function"


@patch("spark_expectations.notifications.plugins.email.Environment")
@patch("spark_expectations.notifications.plugins.email.SparkExpectationsContext", autospec=True, spec_set=True)
def test_process_message_with_empty_message(_mock_context, mock_env):
    """Test that _process_message handles empty messages gracefully"""
    email_handler = SparkExpectationsEmailPluginImpl()
    _mock_context.get_enable_templated_basic_email_body = True
    _mock_context.get_basic_default_template = "Template string"
    _mock_context.get_mail_subject = "Subject"

    # Setup template mock
    mock_template = MagicMock()
    mock_template.module.render_table = "render_function"
    mock_template.render.return_value = "<p>Empty Content</p>"
    mock_env.return_value.from_string.return_value = mock_template

    # Test with empty message
    config_args = {
        "message": "",
        "content_type": "plain",
    }

    mail_content, content_type = email_handler._process_message(_mock_context, config_args)

    # Check that template was rendered with empty title and rows
    call_args = mock_template.render.call_args[1]
    assert call_args["title"] == ""
    assert call_args["rows"] == []
    assert content_type == "html"


@patch("spark_expectations.notifications.plugins.email.Environment")
@patch("spark_expectations.notifications.plugins.email.SparkExpectationsContext", autospec=True, spec_set=True)
def test_process_message_with_malformed_data(_mock_context, mock_env):
    """Test that message parsing handles malformed data gracefully"""
    email_handler = SparkExpectationsEmailPluginImpl()
    _mock_context.get_enable_templated_basic_email_body = True
    _mock_context.get_basic_default_template = "Template string"
    _mock_context.get_mail_subject = "Subject"

    # Setup template mock
    mock_template = MagicMock()
    mock_template.module.render_table = "render_function"
    mock_template.render.return_value = "<p>Rendered Content</p>"
    mock_env.return_value.from_string.return_value = mock_template

    # Test with malformed message (no key-value pairs)
    test_message = "Just a single line with no key-value pairs"

    config_args = {
        "message": test_message,
        "content_type": "plain",
    }

    mail_content, content_type = email_handler._process_message(_mock_context, config_args)

    # Check that title was extracted and empty rows were passed
    call_args = mock_template.render.call_args[1]
    assert call_args["title"] == "Just a single line with no key-value pairs"
    assert call_args["rows"] == []


@patch("spark_expectations.notifications.plugins.email.SparkExpectationsContext", autospec=True, spec_set=True)
def test_process_message_no_content_type(_mock_context):
    """Test that _process_message defaults to plain text when no content_type is provided"""
    email_handler = SparkExpectationsEmailPluginImpl()
    _mock_context.get_enable_templated_basic_email_body = False

    # No content_type provided
    config_args = {"message": "Test message without content type"}

    mail_content, content_type = email_handler._process_message(_mock_context, config_args)

    assert mail_content == "Test message without content type"
    assert content_type == "plain"


# test default template when type is custom
@patch("spark_expectations.notifications.plugins.email.Environment")
@patch("spark_expectations.notifications.plugins.email.PackageLoader")
@patch("spark_expectations.notifications.plugins.email.SparkExpectationsContext", autospec=True, spec_set=True)
def test_process_message_with_custom_template(_mock_context, mock_fs_loader, mock_env):
    """Test that default custom template is used"""
    email_handler = SparkExpectationsEmailPluginImpl()
    _mock_context.get_enable_templated_custom_email = True
    _mock_context.get_custom_default_template = None
    _mock_context.get_mail_subject = "Test Subject"

    # Setup template mocks
    mock_template = MagicMock()
    mock_template.module.render_table = "render_function"
    mock_template.render.return_value = "<table>Rendered HTML</table>"
    mock_env.return_value.get_template.return_value = mock_template

    # Test message with format similar to your expected format
    test_message = 'CUSTOM EMAIL\n{"product_id": "product_id1", "table_name": "test_table"}'

    config_args = {
        "message": test_message,
        "content_type": "plain",
    }

    mail_content, content_type = email_handler._process_message(_mock_context, config_args)

    # Check that the template was used
    mock_fs_loader.assert_called_once_with("spark_expectations","config/templates")
    mock_env.assert_called_once_with(loader=mock_fs_loader.return_value)
    mock_env.return_value.get_template.assert_called_once_with("custom_email_alert_template.jinja")
    mock_template.render.assert_called_once()

    # Check the resulting content
    assert "<table>Rendered HTML</table>" in mail_content
    assert content_type == "html"

    # Verify the message data was parsed correctly
    call_args = mock_template.render.call_args[0]
    assert call_args[0] == {'product_id': 'product_id1', 'table_name': 'test_table'}


# test template from user config when type is custom
@patch("spark_expectations.notifications.plugins.email.Environment")
@patch("spark_expectations.notifications.plugins.email.BaseLoader")
@patch("spark_expectations.notifications.plugins.email.SparkExpectationsContext", autospec=True, spec_set=True)
def test_process_message_with_custom_template(_mock_context, mock_base_loader, mock_env):
    """Test custom email template processing when a template is provided in the user config"""
    email_handler = SparkExpectationsEmailPluginImpl()
    _mock_context.get_enable_templated_custom_email = True
    _mock_context.get_custom_default_template = "Custom template"
    _mock_context.get_mail_subject = "Test Subject"

    # Setup template mocks
    mock_template = MagicMock()
    mock_template.module.render_table = "render_function"
    mock_template.render.return_value = "<div>Custom Template</div>"
    mock_env.return_value.from_string.return_value = mock_template

    # Test message with format similar to your expected format
    test_message = 'CUSTOM EMAIL\n{"product_id": "product_id1", "table_name": "test_table"}'

    config_args = {
        "message": test_message,
        "content_type": "plain",
    }

    mail_content, content_type = email_handler._process_message(_mock_context, config_args)

    # Check the template was used with BaseLoader
    mock_env.return_value.from_string.assert_called_with("Custom template")
    mock_template.render.assert_called_once()

    # Check the resulting content
    assert "<div>Custom Template</div>" in mail_content
    assert content_type == "html"

    # Verify the message data was parsed correctly
    call_args = mock_template.render.call_args[0]
    assert call_args[0] == {'product_id': 'product_id1', 'table_name': 'test_table'}

# test custom email throws json error
@patch("spark_expectations.notifications.plugins.email.Environment")
@patch("spark_expectations.notifications.plugins.email.PackageLoader")
@patch("spark_expectations.notifications.plugins.email._log")
def test_process_message_invalid_json_logs_and_fallback(mock_log, mock_fs_loader, mock_env):
    email_handler = SparkExpectationsEmailPluginImpl()
    context = MagicMock()
    context.get_enable_templated_custom_email = True
    context.get_custom_default_template = None

    # Simulate invalid JSON in mail_content
    config_args = {
        "message": "CUSTOM EMAIL\n{'invalid': unquoted_value}",  # Not valid JSON
        "content_type": "plain"
    }

    mail_content, content_type = email_handler._process_message(context, config_args)

    # Check that the error was logged
    assert mock_log.error.call_count == 1
    assert "JSON decode error" in mock_log.error.call_args[0][0]

    # Check that the returned content is a plain text error message
    assert mail_content.startswith("Error: Invalid JSON format in custom email content.")
    assert "Original content:" in mail_content
    assert content_type == "plain"

# test custom email template rendering throws error
@patch("spark_expectations.notifications.plugins.email.Environment")
@patch("spark_expectations.notifications.plugins.email.PackageLoader")
@patch("spark_expectations.notifications.plugins.email._log")
def test_process_message_template_render_exception(mock_log, mock_fs_loader, mock_env):
    email_handler = SparkExpectationsEmailPluginImpl()
    context = MagicMock()
    context.get_enable_templated_custom_email = True
    context.get_custom_default_template = None

    # Valid JSON, but template.render will raise a general Exception
    config_args = {
        "message": 'CUSTOM EMAIL\n{"product_id": "product_id1", "table_name": "test_table"}',
        "content_type": "plain"
    }

    # Setup template mock to raise Exception on render
    mock_template = MagicMock()
    mock_template.render.side_effect = Exception("Render error")
    mock_env.return_value.get_template.return_value = mock_template

    mail_content, content_type = email_handler._process_message(context, config_args)

    # Check that the error was logged
    assert mock_log.error.call_count >= 1
    assert "Template rendering error" in mock_log.error.call_args[0][0]

    # Check that the returned content is a plain text error message
    assert mail_content.startswith("Error: Failed to render custom email template.")
    assert "Original content:" in mail_content
    assert content_type == "plain"



import logging
import pytest
import yaml
from unittest.mock import Mock, patch, mock_open
from spark_expectations import get_default_log_handler, setup_logger


@pytest.fixture(name="_fixture_logger_handler")
def fixture_logger_handler():
    return get_default_log_handler()


def test_logger_handler_type(_fixture_logger_handler):
    assert isinstance(_fixture_logger_handler, logging.Handler)


def test_logger_handler_formatter(_fixture_logger_handler):
    formatter = _fixture_logger_handler.formatter
    assert isinstance(formatter, logging.Formatter)
    assert (
        formatter._fmt == "[%(asctime)s] [%(levelname)s] [spark_expectations]"
        " {%(module)s.py:%(funcName)s:%(lineno)d} - %(message)s"
    )


@pytest.mark.parametrize("pkg_name", ["foo", "bar"])
def test_logger_handler_pkg_name(pkg_name):
    logger_handler = get_default_log_handler(pkg_name)
    formatter = logger_handler.formatter
    assert pkg_name in formatter._fmt


@pytest.fixture(name="_fixture_logger_name")
def fixture_logger_name():
    return "test_logger"


@pytest.fixture(name="_fixture_logger")
def fixture_logger(_fixture_logger_name):
    return setup_logger(_fixture_logger_name)


@pytest.mark.parametrize("name", [None, "custom_name"])
def test_setup_logger(name, _fixture_logger_name, _fixture_logger):
    log = setup_logger(name)
    assert log.name == name or _fixture_logger_name
    assert log.level == logging.INFO
    assert log.propagate is False
    assert len(log.handlers) == 1


def test_infer_safe_cast_comprehensive_coverage():
    """Test infer_safe_cast to cover all missing branches and improve coverage"""
    from spark_expectations.core import infer_safe_cast
    
    # Test valid dictionary parsing that should hit line 220 (return parsed)
    valid_dict_str = "{'key': 'value', 'number': 42}"
    result_dict = infer_safe_cast(valid_dict_str)
    expected_dict = {'key': 'value', 'number': 42}
    assert result_dict == expected_dict
    assert isinstance(result_dict, dict)
    
    # Test another valid dictionary
    valid_dict_str2 = "{'nested': {'inner': 'value'}}"
    result_dict2 = infer_safe_cast(valid_dict_str2)
    assert result_dict2 == {'nested': {'inner': 'value'}}
    assert isinstance(result_dict2, dict)
    
    # Test malformed dictionary string that triggers ValueError/SyntaxError (line 223)
    malformed_dict = "{'key': invalid_syntax without quotes}"
    result = infer_safe_cast(malformed_dict)
    assert result == malformed_dict  # Should return original string when dict parsing fails
    
    # Test another malformed dictionary that triggers SyntaxError
    malformed_dict2 = "{'unclosed_dict': 'value'"
    result2 = infer_safe_cast(malformed_dict2)
    assert result2 == malformed_dict2  # Should return original string
    
    # Test edge cases for comprehensive coverage
    assert infer_safe_cast(None) is None
    assert infer_safe_cast("") == ""
    assert infer_safe_cast(" ") == ""
    assert infer_safe_cast("  null  ") is None
    assert infer_safe_cast("TRUE") is True
    assert infer_safe_cast("FALSE") is False


def test_load_configurations_error_handling():
    """Test load_configurations error handling - covers lines 50-55"""
    from spark_expectations.core import load_configurations
    import tempfile
    import os
    
    mock_spark = Mock()
    mock_spark.conf.set = Mock()
    
    # Test FileNotFoundError (line 50)
    with patch('builtins.open', side_effect=FileNotFoundError("Config file not found")):
        with pytest.raises(RuntimeError, match="Spark config YAML file not found"):
            load_configurations(mock_spark)
    
    # Test YAMLError (line 52)  
    with patch('builtins.open', mock_open(read_data="invalid: yaml: content: [")):
        with patch('yaml.safe_load', side_effect=yaml.YAMLError("Invalid YAML")):
            with pytest.raises(RuntimeError, match="Error parsing Spark config YAML configuration file"):
                load_configurations(mock_spark)
    
    # Test general Exception (line 54)
    with patch('builtins.open', side_effect=Exception("Unexpected error")):
        with pytest.raises(RuntimeError, match="An unexpected error occurred while loading spark configurations"):
            load_configurations(mock_spark)


def test_get_config_dict_error_conditions():
    """Test get_config_dict error handling"""
    from spark_expectations.core import get_config_dict
    
    mock_spark = Mock()
    
    # Test general Exception
    with patch('spark_expectations.core.load_configurations', side_effect=Exception("Config error")):
        user_conf = {"spark.expectations.is.serverless": False}
        
        with pytest.raises(RuntimeError, match="Error retrieving configuration"):
            get_config_dict(mock_spark, user_conf)


def test_load_configurations_success_scenarios():
    """Test load_configurations with valid config data - covers lines 33-47"""
    from spark_expectations.core import load_configurations
    import json
    
    mock_spark = Mock()
    mock_spark.conf.set = Mock()
    
    # Test with None config (line 34-35)
    with patch('builtins.open', mock_open(read_data="")):
        with patch('yaml.safe_load', return_value=None):
            load_configurations(mock_spark)
            # Should handle None config gracefully
    
    # Test with invalid config type (line 36-37) 
    with patch('builtins.open', mock_open(read_data="invalid_list_config")):
        with patch('yaml.safe_load', return_value=["invalid", "config"]):
            with pytest.raises(RuntimeError, match="Error parsing Spark config YAML configuration file"):
                load_configurations(mock_spark)
    
    # Test valid config processing (lines 38-47)
    valid_config = {
        "se.streaming.topic.name": "test_topic",
        "spark.expectations.notifications.email.enabled": True,
        "spark.sql.adaptive.enabled": True
    }
    
    with patch('builtins.open', mock_open()):
        with patch('yaml.safe_load', return_value=valid_config):
            streaming_dict, notification_dict = load_configurations(mock_spark)
            
            # Verify config processing
            mock_spark.conf.set.assert_any_call("spark.sql.adaptive.enabled", "True")
            # Verify the returned dictionaries
            assert streaming_dict == {"se.streaming.topic.name": "test_topic"}
            assert notification_dict == {"spark.expectations.notifications.email.enabled": True}


def test_build_config_dict_comprehensive():
    """Test _build_config_dict internal function comprehensively"""
    from spark_expectations.core import get_config_dict
    
    mock_spark = Mock()
    # Mock spark.conf.get to return the default value (second argument) when called
    mock_spark.conf.get = Mock(side_effect=lambda key, default=None: default)
    
    # Test with user_conf that overrides defaults
    with patch('spark_expectations.core.load_configurations') as mock_load_configs:
        mock_load_configs.return_value = (
            {},
            {"spark.expectations.notifications.email.enabled": True}
        )
        user_conf_with_overrides = {
            "spark.expectations.notifications.email.enabled": False,
            "custom.config": "test_value"
        }
        
        notification_dict, streaming_dict = get_config_dict(mock_spark, user_conf_with_overrides)
        
        # Verify user_conf overrides defaults
        assert notification_dict["spark.expectations.notifications.email.enabled"] == False
        assert streaming_dict == {}  # Empty from load_configurations
    
    # Test with user_conf containing partial configs
    # Note: _build_config_dict only includes keys from default_dict, not from user_conf
    with patch('spark_expectations.core.load_configurations') as mock_load_configs:
        mock_load_configs.return_value = (
            {"se.streaming.enable": False},
            {
                "spark.expectations.notifications.email.enabled": True,
                "spark.expectations.notifications.slack.enabled": False
            }
        )
        
        user_conf_partial = {
            "spark.expectations.notifications.slack.enabled": True
        }
        
        notification_dict, streaming_dict = get_config_dict(mock_spark, user_conf_partial)
        
        # Verify user_conf partial values merge with defaults
        # email.enabled is not in user_conf, so it gets default value from load_configurations
        assert notification_dict["spark.expectations.notifications.email.enabled"] == True  # from defaults
        # slack.enabled is in user_conf, so it overrides the default
        assert notification_dict["spark.expectations.notifications.slack.enabled"] == True  # from user_conf
        assert streaming_dict["se.streaming.enable"] == False  # from defaults
    
    # Test user_conf is None - should use defaults directly
    with patch('spark_expectations.core.load_configurations') as mock_load_configs:
        mock_load_configs.return_value = (
            {},
            {"spark.expectations.notifications.email.enabled": False}
        )
        
        notification_dict, streaming_dict = get_config_dict(mock_spark, None)
        
        # Verify None user_conf uses defaults
        assert notification_dict["spark.expectations.notifications.email.enabled"] == False
        assert isinstance(notification_dict, dict)
        assert isinstance(streaming_dict, dict)


def test_build_config_dict_no_user_conf_uses_defaults():
    """Test _build_config_dict when user_conf is None - should use defaults directly"""
    from spark_expectations.core import get_config_dict
    
    mock_spark = Mock()
    # Mock spark.conf.get to return the default value (second argument) when called
    mock_spark.conf.get = Mock(side_effect=lambda key, default=None: default)
    
    # Mock scenario where user_conf is None and we use defaults
    with patch('spark_expectations.core.load_configurations') as mock_load_configs:
        mock_load_configs.return_value = (
            {"se.streaming.enable": True},
            {"spark.expectations.notifications.email.enabled": False}
        )
        
        # When user_conf is None, should use defaults directly
        notification_dict, streaming_dict = get_config_dict(mock_spark, None)
        
        # Verify defaults are used
        assert notification_dict["spark.expectations.notifications.email.enabled"] == False
        assert streaming_dict["se.streaming.enable"] == True
        assert isinstance(notification_dict, dict)
        assert isinstance(streaming_dict, dict)


def test_get_spark_session_local_environment():
    """Test get_spark_session with UNIT_TESTING_ENV set - covers lines 143-168"""
    from spark_expectations.core import get_spark_session
    import os
    
    # Test UNIT_TESTING_ENV branch (lines 143-168)
    with patch.dict(os.environ, {"UNIT_TESTING_ENV": "spark_expectations_unit_testing_on_github_actions"}):
        with patch('spark_expectations.core.SparkSession.builder') as mock_builder:
            mock_session = Mock()
            mock_chain = Mock()
            
            # Setup the builder chain
            mock_builder.config.return_value = mock_chain
            mock_chain.config.return_value = mock_chain
            mock_chain.getOrCreate.return_value = mock_session
            
            with patch('spark_expectations.core.load_configurations') as mock_load_config:
                result = get_spark_session()
                
                # Verify the SparkSession builder was configured with Delta Lake settings
                mock_builder.config.assert_called_with("spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension")
                # Verify load_configurations was called
                mock_load_config.assert_called_once_with(mock_session)
                assert result == mock_session


def test_get_spark_session_sparkexpectations_env():
    """Test get_spark_session with SPARKEXPECTATIONS_ENV set - covers lines 143-168"""
    from spark_expectations.core import get_spark_session
    import os
    
    # Test SPARKEXPECTATIONS_ENV branch (lines 143-168)
    with patch.dict(os.environ, {"SPARKEXPECTATIONS_ENV": "local"}, clear=True):
        with patch('spark_expectations.core.SparkSession.builder') as mock_builder:
            mock_session = Mock()
            mock_chain = Mock()
            
            # Setup the builder chain
            mock_builder.config.return_value = mock_chain
            mock_chain.config.return_value = mock_chain
            mock_chain.getOrCreate.return_value = mock_session
            
            with patch('spark_expectations.core.load_configurations') as mock_load_config:
                result = get_spark_session()
                
                # Verify JAR configuration was set (lines 154-160)
                jar_calls = [call for call in mock_chain.config.call_args_list if 'spark.jars' in str(call)]
                assert len(jar_calls) > 0  # Verify JAR config was called
                
                # Verify other configurations (lines 161-165)
                config_calls = [str(call) for call in mock_chain.config.call_args_list]
                assert any('spark.sql.shuffle.partitions' in call for call in config_calls)
                assert any('spark.ui.enabled' in call for call in config_calls)
                
                mock_load_config.assert_called_once_with(mock_session)
                assert result == mock_session


def test_get_spark_session_active_session_branch():
    """Test get_spark_session else branch using getActiveSession - covers lines 169-171"""
    from spark_expectations.core import get_spark_session
    import os
    
    # Test else branch (lines 169-171)
    with patch.dict(os.environ, {}, clear=True):  # Clear environment variables
        with patch('spark_expectations.core.SparkSession.getActiveSession') as mock_get_active:
            mock_session = Mock()
            mock_get_active.return_value = mock_session
            
            with patch('spark_expectations.core.load_configurations') as mock_load_config:
                result = get_spark_session()
                
                # Verify getActiveSession was called (line 170)
                mock_get_active.assert_called_once()
                mock_load_config.assert_called_once_with(mock_session)
                assert result == mock_session


def test_infer_safe_cast_string_fallback():
    """Test infer_safe_cast final string fallback - covers lines 219-220"""
    from spark_expectations.core import infer_safe_cast
    
    # Test final fallback to string (lines 219-220)
    # Use strings that can't be converted to int, float, bool, or dict
    test_strings = [
        "this is just a regular string that can't be converted",
        "not_a_number_or_bool",
        "random_text_123_abc",
        "hello world!",
        "some-special-chars@#$%"
    ]
    
    for test_str in test_strings:
        result = infer_safe_cast(test_str)
        # Should return the cleaned string (line 220)
        assert result == test_str
        assert isinstance(result, str)
    
    # Test with string that has whitespace - should be stripped
    whitespace_string = "   some string with spaces   "
    result2 = infer_safe_cast(whitespace_string)
    assert result2 == "some string with spaces"  # Stripped
    
    # Test with non-string input that falls back to string
    class CustomObject:
        def __str__(self):
            return "custom_object_string"
    
    custom_obj = CustomObject()
    result3 = infer_safe_cast(custom_obj)
    assert result3 == "custom_object_string"
    assert isinstance(result3, str)
    
    # Test with complex object that becomes string
    test_tuple = (1, 2, 3)
    result4 = infer_safe_cast(test_tuple)
    assert result4 == "(1, 2, 3)"
    assert isinstance(result4, str)


def test_infer_safe_cast_complete_fallback_coverage():
    """Ensure we hit all branches including the dictionary parsing exception - covers lines 220-225"""
    from spark_expectations.core import infer_safe_cast
    
    # Test various inputs that will definitely hit the final fallback
    test_cases = [
        "definitely_not_a_number",
        "not_true_or_false", 
        "not{a:valid}dict",
        "random text with symbols !@#$%^&*()",
        "   whitespace_test   ",  # This will be stripped
        "{'incomplete': dict without closing brace",  # Should trigger SyntaxError in ast.literal_eval
        "[1, 2, 3",  # Another syntax error case
        "{'key': 'value'} extra text",  # Valid dict but with extra text
    ]
    
    for test_input in test_cases:
        result = infer_safe_cast(test_input)
        expected = str(test_input).strip()
        assert result == expected
        assert isinstance(result, str)
        
    # Test with a bytes object to ensure it converts to string
    bytes_input = b"bytes_string"
    result_bytes = infer_safe_cast(bytes_input)
    assert result_bytes == "b'bytes_string'"
    assert isinstance(result_bytes, str)
    
    # Specifically test the dictionary parsing exception path
    malformed_dict_cases = [
        "{'key': value_without_quotes}",  # Should trigger ValueError in ast.literal_eval
        "{key: 'value'}",  # Invalid dict format 
        "{'nested': {'incomplete': }",  # Nested incomplete dict
    ]
    
    for malformed_dict in malformed_dict_cases:
        result = infer_safe_cast(malformed_dict)
        # Should fall back to string due to ValueError/SyntaxError in dict parsing
        assert result == malformed_dict
        assert isinstance(result, str)



from spark_expectations.config.user_config import Constants as user_config


def test_constants():
    assert user_config.se_dq_obs_alert_flag == "spark.expectations.notifications.alert.flag.disable"

    assert user_config.se_notifications_email_smtp_port == "spark.expectations.notifications.email.smtp.port"

    assert user_config.se_notifications_email_smtp_port == "spark.expectations.notifications.email.smtp.port"

    assert user_config.se_notifications_email_from == "spark.expectations.notifications.email.from"

    assert (
        user_config.se_notifications_email_to_other_mail_id == "spark.expectations.notifications.email."
        "to.other.mail.com"
    )

    assert user_config.se_notifications_email_subject == "spark.expectations.notifications.email.subject"

    assert user_config.se_notifications_enable_slack == "spark.expectations.notifications.slack.enabled"

    assert user_config.se_notifications_slack_webhook_url == "spark.expectations.notifications.slack.webhook.url"

    assert user_config.se_notifications_enable_zoom == "spark.expectations.notifications.zoom.enabled"

    assert user_config.se_notifications_zoom_webhook_url == "spark.expectations.notifications.zoom.webhook.url"

    assert user_config.se_notifications_zoom_token == "spark.expectations.notifications.zoom.token"

    assert user_config.se_notifications_enable_pagerduty == "spark.expectations.notifications.pagerduty.enabled"

    assert user_config.se_notifications_pagerduty_integration_key == "spark.expectations.notifications.pagerduty.integration.key"
    
    assert user_config.se_notifications_pagerduty_webhook_url == "spark.expectations.notifications.pagerduty.webhook.url"

    assert user_config.se_notifications_on_start == "spark.expectations.notifications.on.start"

    assert user_config.se_notifications_on_completion == "spark.expectations.notifications.on.completion"

    assert user_config.se_notifications_on_fail == "spark.expectations.notifications.on.fail"

    assert (
        user_config.se_notifications_on_error_drop_exceeds_threshold_breach == "spark.expectations."
        "notifications.on.error.drop."
        "exceeds.threshold.breach"
    )

    assert (
        user_config.se_notifications_on_error_drop_threshold == "spark.expectations.notifications."
        "error.drop.threshold"
    )

    assert user_config.se_enable_streaming == "se.streaming.enable"

    assert user_config.se_enable_error_table == "se.enable.error.table"

    assert user_config.se_dq_rules_params == "se.dq.rules.params"

    assert user_config.secret_type == "se.streaming.secret.type"

    assert user_config.cbs_url == "se.streaming.cerberus.url"
    assert user_config.cbs_sdb_path == "se.streaming.cerberus.sdb.path"
    assert user_config.cbs_kafka_server_url == "se.streaming.cerberus.kafka.server.url"
    assert user_config.cbs_secret_token_url == "se.streaming.cbs.secret.token.url"
    assert user_config.cbs_secret_app_name == "se.streaming.cbs.secret.app.name"
    assert user_config.cbs_secret_token == "se.streaming.cerberus.secret.token"
    assert user_config.cbs_topic_name == "se.streaming.cerberus.token.name"

    assert user_config.dbx_workspace_url == "se.streaming.dbx.workspace.url"
    assert user_config.dbx_secret_scope == "se.streaming.dbx.secret.scope"
    assert user_config.dbx_kafka_server_url == "se.streaming.dbx.kafka.server.url"
    assert user_config.dbx_secret_token_url == "se.streaming.dbx.secret.token.url"
    assert user_config.dbx_secret_app_name == "se.streaming.dbx.secret.app.name"
    assert user_config.dbx_secret_token == "se.streaming.dbx.secret.token"
    assert user_config.dbx_topic_name == "se.streaming.dbx.topic.name"

    assert user_config.se_enable_agg_dq_detailed_result == "spark.expectations.agg.dq.detailed.stats"
    assert user_config.se_enable_query_dq_detailed_result == "spark.expectations.query.dq.detailed.stats"

    assert user_config.se_dq_obs_default_email_template == "spark.expectations.dq.obs.default.detailed.email.template"
    assert user_config.se_notifications_default_basic_email_template == "spark.expectations.notifications.email.default.basic.template"
    assert user_config.se_notifications_email_custom_body == "spark.expectations.notifications.email.custom.body"
    assert user_config.se_notifications_enable_templated_basic_email_body == (
        "spark.expectations.notifications.email.templated.basic.body.enable"
    ) 
    assert user_config.se_notifications_enable_custom_email_body == "spark.expectations.notifications.email.custom.body.enable"
    assert user_config.se_notifications_enable_templated_custom_email == (
        "spark.expectations.notifications.email.templated.custom.email.enable"
    )
    assert user_config.se_notifications_email_custom_template == "spark.expectations.notifications.email.custom.template"
